import logging
import inspect

import yaml

from .declerror import DeclarationError
from .abc import Expression

from .expression.value.valueexpr import VALUE
from .expression.statement.funexpr import FUNCTION

from .expression.utility.castexpr import CAST
from .expression.statement.selfexpr import SELF

###

L = logging.getLogger(__name__)

###


class IncludeNeeded(Exception):
    def __init__(self, identifier):
        super().__init__()
        self.Identifier = identifier


class ExpressionBuilder(object):
    """
    Builds an expression from configuration.
    """

    def __init__(self, app, library=None, include_paths=None):
        self.App = app
        self.ExpressionClasses = {}
        self.Identifier = None

        self.Config = {}
        self.Library = library

        if include_paths is None:
            self.IncludePaths = ["/include"]

        else:
            self.IncludePaths = include_paths
            assert isinstance(include_paths, list)

        # Cache for loaded includes during the parsing
        self.LoadedIncludes = {}

        # Register the common expression module
        from . import expression

        self.register_module(expression)

    def register_module(self, module):
        for class_name, expression_class in inspect.getmembers(module, inspect.isclass):
            self.register_class(class_name, expression_class)

    def register_class(self, class_name, expression_class):
        class_name = class_name.replace("_", ".")
        self.ExpressionClasses[class_name] = expression_class

    def add_config_value(self, key, value):
        self.Config[key] = value

    def update_config(self, config):
        self.Config.update(config)

    async def read(self, identifier):
        if self.Library is None:
            raise RuntimeError(
                "Cannot read '{}' in builder, ASAB library is not provided".format(
                    identifier
                )
            )

        # Read declaration from available declarations libraries
        for include_path in self.IncludePaths:
            declaration = await self.Library.read(
                "{}/{}.yaml".format(include_path, identifier)
            )

            if declaration is not None:
                return declaration.read().decode("utf-8")

        raise RuntimeError(
            "Cannot find '{}' YAML declaration in libraries".format(identifier)
        )

    async def parse(self, declaration, source_name=None):
        """
        Returns a list of expressions from the loaded declaration.
        :param declaration:
        :param source_name:
        :return:
        """
        self.Identifier = None

        if isinstance(declaration, str) and declaration.startswith("---"):
            pass

        else:
            self.Identifier = declaration
            declaration = await self.read(declaration)

        while True:
            loader = yaml.Loader(declaration)
            if source_name is not None:
                loader.name = source_name

            # Register the constructor for each registered expression class
            for name in self.ExpressionClasses:
                loader.add_constructor("!{}".format(name), self._constructor)

            loader.add_constructor("!INCLUDE", self._construct_include)
            loader.add_constructor("!CONFIG", self._construct_config)

            loader.add_constructor("tag:yaml.org,2002:ui256", self._construct_scalar)
            loader.add_constructor("tag:yaml.org,2002:ui128", self._construct_scalar)
            loader.add_constructor("tag:yaml.org,2002:ui64", self._construct_scalar)
            loader.add_constructor("tag:yaml.org,2002:ui32", self._construct_scalar)
            loader.add_constructor("tag:yaml.org,2002:ui16", self._construct_scalar)
            loader.add_constructor("tag:yaml.org,2002:ui8", self._construct_scalar)

            loader.add_constructor("tag:yaml.org,2002:si256", self._construct_scalar)
            loader.add_constructor("tag:yaml.org,2002:si128", self._construct_scalar)
            loader.add_constructor("tag:yaml.org,2002:si64", self._construct_scalar)
            loader.add_constructor("tag:yaml.org,2002:si32", self._construct_scalar)
            loader.add_constructor("tag:yaml.org,2002:si16", self._construct_scalar)
            loader.add_constructor("tag:yaml.org,2002:si8", self._construct_scalar)

            loader.add_constructor("tag:yaml.org,2002:fp128", self._construct_scalar)
            loader.add_constructor("tag:yaml.org,2002:fp64", self._construct_scalar)
            loader.add_constructor("tag:yaml.org,2002:fp32", self._construct_scalar)
            loader.add_constructor("tag:yaml.org,2002:fp16", self._construct_scalar)

            loader.add_constructor("tag:yaml.org,2002:str", self._construct_scalar)

            try:
                expressions = []

                # Build syntax trees for each expression
                while loader.check_data():
                    expression = loader.get_data()

                    # Run initialize for the expression and any instance inside
                    for parent, key, obj in self._walk(expression):
                        if not isinstance(obj, Expression):
                            continue

                        if source_name != "<INCLUDE>":
                            if isinstance(obj, SELF):
                                # Implement a self-reference or Y-Combinator
                                obj.initialize(expression)
                            else:
                                obj.initialize()

                    expressions.append(expression)

            except yaml.scanner.ScannerError as e:
                raise DeclarationError("Syntax error in declaration: {}".format(e))

            except yaml.constructor.ConstructorError as e:
                raise DeclarationError("Unknown declarative expression: {}".format(e))

            except IncludeNeeded as e:
                # If include is needed, load its declaration to the loaded include cache
                include_declaration = await self.read(e.Identifier)
                parsed_declaration = await self.parse(include_declaration, "<INCLUDE>")

                # Include can be only one expression
                self.LoadedIncludes[e.Identifier] = parsed_declaration[0]
                continue

            finally:
                loader.dispose()

            return expressions

    async def parse_ext(self, declaration, source_name=None):
        """
        Wrap top-level declaration into a function, value etc.
        This is likely intermediate (not a final) implementation.
        """
        result = []
        for expr in await self.parse(declaration, source_name=source_name):
            if isinstance(expr, (VALUE, FUNCTION)):
                result.append(expr)
                continue

            if isinstance(expr, Expression):
                fun_expr = FUNCTION(self.App, arg_apply=expr, arg_name="main")
                fun_expr.set_location(expr.Location)
                result.append(fun_expr)
                continue

            if isinstance(expr, (bool, int, float, str, set, list, dict)):
                expr = VALUE(self.App, value=expr)
                # Fake the location string
                expr.set_location(
                    "  at line 1, column 1:\n{}\n^".format(declaration[:20])
                )
                result.append(expr)
                continue

            raise NotImplementedError(
                "Top-level type '{}' not supported".format(type(expr))
            )

        return result

    def _walk(self, expression):
        if expression is None:
            return

        if isinstance(expression, Expression):
            for parent, key, obj in expression.walk():
                yield (parent, key, obj)

                if isinstance(obj, (dict, list)):
                    for _parent, _key, _obj in self._walk(obj):
                        yield (_parent, _key, _obj)

        elif isinstance(expression, dict):
            for _key, _expression in expression.items():
                for parent, key, obj in self._walk(_expression):
                    yield (parent, key, obj)

        elif isinstance(expression, (list, set, frozenset)):
            for _expression in expression:
                for parent, key, obj in self._walk(_expression):
                    yield (parent, key, obj)

        elif isinstance(expression, (str, int, float, tuple)):
            return

        else:
            raise NotImplementedError(
                "Walk not implemented for '{}'.".format(expression)
            )

    def _construct_include(self, loader: yaml.Loader, node: yaml.Node):
        """Include file referenced at node."""

        identifier = loader.construct_scalar(node)

        try:
            return self.LoadedIncludes[identifier]

        except KeyError:
            raise IncludeNeeded(identifier)

    def _construct_config(self, loader: yaml.Loader, node: yaml.Node):
        key = loader.construct_scalar(node)
        return self.Config.get(key)

    def _constructor(self, loader, node):
        assert node.tag[0] == "!"
        xclass = self.ExpressionClasses[node.tag[1:]]

        location = node.start_mark

        if self.Identifier is not None:
            # https://github.com/yaml/pyyaml/blob/4c2e993321ad29a02a61c4818f3cef9229219003/lib3/yaml/reader.py
            location = location.replace("<unicode string>", str(self.Identifier))

        try:
            if isinstance(node, yaml.ScalarNode):
                value = loader.construct_scalar(node)
                obj = xclass(app=self.App, value=value)

            elif isinstance(node, yaml.SequenceNode):
                value = loader.construct_sequence(node)
                obj = xclass(app=self.App, sequence=value)

            elif isinstance(node, yaml.MappingNode):
                value = loader.construct_mapping(node)
                obj = xclass(
                    app=self.App, **dict(("arg_" + k, v) for k, v in value.items())
                )

            else:
                raise RuntimeError("Unsupported node type '{}'".format(node))

            obj.set_location(location)
            obj.Node = node
            return obj

        except IncludeNeeded as e:
            raise e

        except TypeError as e:
            raise DeclarationError("Type error {}\n{}\n".format(e, node.start_mark))

        except Exception as e:
            L.exception("Error in expression")
            raise DeclarationError(
                "Invalid expression at {}\n{}".format(node.start_mark, e)
            )

    def _construct_scalar(self, loader: yaml.Loader, node: yaml.Node):
        location = node.start_mark
        if self.Identifier is not None:
            # https://github.com/yaml/pyyaml/blob/4c2e993321ad29a02a61c4818f3cef9229219003/lib3/yaml/reader.py
            location = location.replace("<unicode string>", str(self.Identifier))

        if isinstance(node, yaml.ScalarNode):
            # Value variant e.g.: `!!si 64`
            outlet_type = node.tag.rsplit(":", 1)[1]
            if outlet_type == "str":
                return node.value
            elif outlet_type[0] in ("u", "s"):
                value = int(node.value)
            elif outlet_type[0] == "f":
                value = float(node.value)
            else:
                raise NotImplementedError(
                    "Unsupport scalar type '{}'".format(outlet_type)
                )
            obj = VALUE(self.App, value=value, outlet_type=outlet_type)

        elif isinstance(node, yaml.MappingNode):
            # Casting variant
            outlet_type = node.tag.rsplit(":", 1)[1]
            value = loader.construct_mapping(node)
            obj = CAST(self.App, arg_what=value["what"], arg_type=outlet_type)

        else:
            NotImplementedError("Unimplemented for YAML node '{}'".format(node))

        obj.set_location(location)
        obj.Node = node
        return obj



import os
import io
import typing
import hashlib
import logging
import tempfile
import dataclasses
import urllib.parse
import xml.dom.minidom

import aiohttp

from ...config import Config
from ..item import LibraryItem
from .abc import LibraryProviderABC

#

L = logging.getLogger(__name__)


#


class AzureStorageLibraryProvider(LibraryProviderABC):
    """
    AzureStorageLibraryProvider is a library provider that reads
    from an Microsoft Azure Storage container.

    Configure by:

    azure+https://ACCOUNT-NAME.blob.core.windows.net/BLOB-CONTAINER

    If Container Public Access Level is not set to "Public access",
    then "Access Policy" must be created with "Read" and "List" permissions
    and "Shared Access Signature" (SAS) query string must be added to a URL in a configuration:

    azure+https://ACCOUNT-NAME.blob.core.windows.net/BLOB-CONTAINER?sv=2020-10-02&si=XXXX&sr=c&sig=XXXXXXXXXXXXXX

    """

    def __init__(self, library, path, layer):
        super().__init__(library, layer)
        assert path[:6] == "azure+"
        self.URL = urllib.parse.urlparse(path[6:])
        self.Model = None  # Will be set by `_load_model` method
        self.Path = path

        self.CacheDir = Config.get("library", "azure_cache")
        if self.CacheDir == "false":
            self.CacheDir = None
        elif self.CacheDir == "true":
            self.CacheDir = os.path.join(
                tempfile.gettempdir(),
                "asab.library.azure.{}".format(
                    hashlib.sha256(path.encode("utf-8")).hexdigest()
                ),
            )

        # Ensure that the case directory exists
        if self.CacheDir is not None:
            try:
                os.makedirs(self.CacheDir)
            except FileExistsError:
                pass  # Cache directory already exists

        self.App.TaskService.schedule(self._start())

    async def _start(self):
        await self._load_model()
        if self.Model is not None:
            await self._set_ready()

    # TODO: Call this periodically
    async def _load_model(self):
        url = urllib.parse.urlunparse(
            urllib.parse.ParseResult(
                scheme=self.URL.scheme,
                netloc=self.URL.netloc,
                path=self.URL.path,
                params="",
                query=self.URL.query + "&restype=container&comp=list",
                fragment="",
            )
        )

        async with aiohttp.ClientSession() as session:
            async with session.get(url) as resp:
                if resp.status == 200:
                    content = await resp.text()
                else:
                    err = await resp.text()
                    L.warning("Failed to list blobs from `{}`:\n{}".format(url, err))
                    return

        model = AzureDirectory("/", sub=dict())

        dom = xml.dom.minidom.parseString(content)
        for blob in dom.getElementsByTagName("Blob"):
            path = get_xml_text(blob.getElementsByTagName("Name"))

            path = path.split("/")
            curmodel = model
            for i in range(len(path) - 1):
                newmodel = curmodel.sub.get(path[i])
                if newmodel is None:
                    curmodel.sub[path[i]] = newmodel = AzureDirectory(
                        name="/" + "/".join(path[: i + 1]), sub=dict()
                    )

                curmodel = newmodel

            curmodel.sub[path[-1]] = AzureItem(name="/" + "/".join(path))

        self.Model = model

        # TODO: If the cache is active, remove items from the cache that:
        # 1) are not in the list
        # 2) their etag differs

        L.info("is connected.", struct_data={"path": self.Path})

    async def list(self, path: str) -> list:
        if self.Model is None:
            L.warning(
                "Azure Storage library provider is not ready. Cannot list {}".format(
                    path
                )
            )
            raise RuntimeError("Not ready")

        assert path[:1] == "/"
        assert (
            "//" not in path
        ), "Directory path cannot contain double slashes (//). Example format: /library/Templates/"
        assert len(path) == 1 or path[-1:] != "/"

        if path == "/":
            pathparts = []
        else:
            pathparts = path.split("/")[1:]

        curmodel = self.Model
        for p in pathparts:
            curmodel = curmodel.sub.get(p)
            if curmodel is None:
                raise KeyError("Not '{}' found".format(path))
            if curmodel.type != "dir":
                raise KeyError("Not '{}' found".format(path))

        items = []
        for i in curmodel.sub.values():
            items.append(
                LibraryItem(
                    name=i.name,
                    type=i.type,
                    layer=self.Layer,
                    providers=[self],
                )
            )

        return items

    async def read(self, path: str) -> typing.IO:
        assert path[:1] == "/"
        assert (
            "//" not in path
        ), "Item path cannot contain double slashes (//). Example format: /library/Templates/item.json"
        assert (
            len(path) == 1 or path[-1:] != "/"
        ), "Item path must end with an extension. Example format: /library/Templates/item.json"

        headers = {}

        pathhash = hashlib.sha256(path.encode("utf-8")).hexdigest()
        cachefname = os.path.join(self.CacheDir, pathhash)
        if self.CacheDir is not None:
            try:
                with open(cachefname + ".etag", "r") as etagf:
                    etag = etagf.read()
                # We found a local cached file with the etag, we will use that in the request
                # if the request returns "304 Not Modified" then we will ship the local version of the file
                headers["If-None-Match"] = etag
            except FileNotFoundError:
                pass

        url = urllib.parse.urlunparse(
            urllib.parse.ParseResult(
                scheme=self.URL.scheme,
                netloc=self.URL.netloc,
                path=self.URL.path + path,
                params="",
                query=self.URL.query,
                fragment="",
            )
        )

        async with aiohttp.ClientSession() as session:
            async with session.get(url, headers=headers) as resp:
                if resp.status == 200:
                    etag = resp.headers.get("ETag")

                    if self.CacheDir is not None and etag is not None:
                        output = open(cachefname, "w+b")

                        with open(cachefname + ".etag", "w") as etagf:
                            etagf.write(etag)

                    else:
                        # Store the response into the temporary file
                        # ... that's to avoid storing the whole (and possibly large) file in the memory
                        output = tempfile.TemporaryFile()

                    async for chunk in resp.content.iter_chunked(
                        16 * io.DEFAULT_BUFFER_SIZE
                    ):
                        output.write(chunk)

                elif (
                    resp.status == 304 and self.CacheDir is not None
                ):  # 304 is Not Modified
                    # The file should be read from cache
                    output = open(cachefname, "r+b")

                else:
                    L.warning(
                        "Failed to get blob:\n{}".format(await resp.text()),
                        struct_data={"status": resp.status},
                    )
                    return None

        # Rewind the file so the reader can start consuming from the beginning
        output.seek(0)
        return output


@dataclasses.dataclass
class AzureDirectory:
    name: str
    sub: dict
    type: str = "dir"


@dataclasses.dataclass
class AzureItem:
    name: str
    type: str = "item"


def get_xml_text(nodelist):
    rc = []
    for node in nodelist:
        for textnode in node.childNodes:
            if textnode.nodeType == textnode.TEXT_NODE:
                rc.append(textnode.data)
    return "".join(rc)



import aiohttp.abc

from ..log import LOG_NOTICE
from .. import Config


class AccessLogger(aiohttp.abc.AbstractAccessLogger):
    def __init__(self, logger, log_format) -> None:
        super().__init__(logger, log_format)
        self.App = logger.App
        self.WebService = self.App.get_service("asab.WebService")
        self.web_metrics_config = Config.getboolean(
            "asab:metrics", "web_requests_metrics", fallback=False
        )

    def log(self, request, response, time):
        struct_data = {
            "I": request.remote,
            "al.m": request.method,
            "al.p": request.path,
            "al.c": response.status,
            "D": time,
        }

        if request.content_length is not None:
            struct_data["al.B"] = request.content_length

        if response.content_length is not None:
            struct_data["al.b"] = response.content_length

        if hasattr(request, "Identity"):
            struct_data["i"] = request.Identity

        agent = request.headers.get("User-Agent")
        if agent is not None:
            struct_data["al.A"] = agent

        xfwd = request.headers.get("X-Forwarded-For")
        if xfwd is not None:
            # TODO: Sanitize xfwd
            # In nginx, use "proxy_set_header X-Forwarded-For $proxy_add_x_forwarded_for;"
            struct_data["Ix"] = xfwd[:128]

        self.logger.log(LOG_NOTICE, "", struct_data=struct_data)

        # Metrics
        if self.web_metrics_config:
            path = request.match_info.get_info().get("formatter")
            if path is None:
                path = request.path

            self.WebService.WebRequestsMetrics.set_metrics(
                time, request.method, path, str(response.status)
            )



from .connection import MySQLConnection
from .source import MySQLSource
from .sink import MySQLSink
from .lookup import MySQLLookup


__all__ = (
    "MySQLConnection",
    "MySQLSource",
    "MySQLSink",
    "MySQLLookup",
)



import io
import os
import os.path
import stat
import glob
import struct
import typing
import logging

from .abc import LibraryProviderABC
from ..item import LibraryItem
from ...timer import Timer

try:
    from .filesystem_inotify import (
        inotify_init,
        inotify_add_watch,
        IN_CREATE,
        IN_ISDIR,
        IN_ALL_EVENTS,
        EVENT_FMT,
        EVENT_SIZE,
        IN_MOVED_TO,
        IN_IGNORED,
    )
except OSError:
    inotify_init = None

#

L = logging.getLogger(__name__)

#


class FileSystemLibraryProvider(LibraryProviderABC):
    def __init__(self, library, path, layer, *, set_ready=True):
        """
        `set_ready` can be used to disable/defer `self._set_ready` call.
        """

        super().__init__(library, layer)
        self.BasePath = os.path.abspath(path)
        while self.BasePath.endswith("/"):
            self.BasePath = self.BasePath[:-1]

        L.info("is connected.", struct_data={"path": path})
        # Filesystem is always ready (or you have a serious problem)
        if set_ready:
            self.App.TaskService.schedule(self._set_ready())

        # Open inotify file descriptor
        if inotify_init is not None:
            init = inotify_init()
            if init == -1:
                L.warning(
                    "Subscribing to library changes in filesystem provider is not available. Inotify was not initialized."
                )
                self.FD = None
            else:
                self.FD = init
                self.App.Loop.add_reader(self.FD, self._on_inotify_read)
                self.AggrTimer = Timer(self.App, self._on_aggr_timer)
        else:
            self.FD = None

        self.AggrEvents = []
        self.WDs = {}

    async def read(self, path: str) -> typing.IO:
        node_path = self.BasePath + path

        # File path must start with '/'
        assert (
            node_path[:1] == "/"
        ), "File path must start with a forward slash (/). For example: /library/Templates/file.json"
        # File path must end with the extension
        assert (
            len(os.path.splitext(node_path)[1]) > 0
        ), "File path must end with an extension. For example: /library/Templates/item.json"
        # File cannot contain '//'
        assert (
            "//" not in node_path
        ), "File path cannot contain double slashes (//). Example format: /library/Templates/item.json"

        try:
            return io.FileIO(node_path, "rb")

        except FileNotFoundError:
            return None

        except IsADirectoryError:
            return None

    async def list(self, path: str) -> list:
        # This list method is completely synchronous, but it should look like asynchronous to make all list methods unified among providers.
        return self._list(path)

    def _list(self, path: str):
        node_path = self.BasePath + path

        # Directory path must start with '/'
        assert (
            node_path[:1] == "/"
        ), "Directory path must start with a forward slash (/). For example: /library/Templates/"
        # Directory path must end with '/'
        assert (
            node_path[-1:] == "/"
        ), "Directory path must end with a forward slash (/). For example: /library/Templates/"
        # Directory cannot contain '//'
        assert (
            "//" not in node_path
        ), "Directory path cannot contain double slashes (//). Example format: /library/Templates/"

        exists = os.access(node_path, os.R_OK) and os.path.isdir(node_path)
        if not exists:
            raise KeyError(" '{}' not found".format(path))

        items = []
        for fname in glob.iglob(os.path.join(node_path, "*")):
            fstat = os.stat(fname)

            assert fname.startswith(self.BasePath)
            fname = fname[len(self.BasePath) :]

            if stat.S_ISREG(fstat.st_mode):
                ftype = "item"
            elif stat.S_ISDIR(fstat.st_mode):
                ftype = "dir"
                fname += "/"
            else:
                ftype = "?"

            # Remove any component that starts with '.'
            if any(x.startswith(".") for x in fname.split("/")):
                continue

            items.append(
                LibraryItem(
                    name=fname,
                    type=ftype,
                    layer=self.Layer,
                    providers=[self],
                )
            )

        return items

    def _on_inotify_read(self):
        data = os.read(self.FD, 64 * 1024)

        pos = 0
        while pos < len(data):
            wd, mask, cookie, namesize = struct.unpack_from(EVENT_FMT, data, pos)
            pos += EVENT_SIZE + namesize
            name = (data[pos - namesize : pos].split(b"\x00", 1)[0]).decode()

            if mask & IN_ISDIR == IN_ISDIR and (
                (mask & IN_CREATE == IN_CREATE) or (mask & IN_MOVED_TO == IN_MOVED_TO)
            ):
                subscribed_path, child_path = self.WDs[wd]
                self._subscribe_recursive(subscribed_path, "/".join([child_path, name]))

            if mask & IN_IGNORED == IN_IGNORED:
                # cleanup
                del self.WDs[wd]
                continue

            self.AggrEvents.append((wd, mask, cookie, os.fsdecode(name)))

        self.AggrTimer.restart(0.2)

    async def _on_aggr_timer(self):
        to_advertise = set()
        for wd, mask, cookie, name in self.AggrEvents:
            # When wathed directory is being removed, more than one inotify events are being produced.
            # When IN_IGNORED event occurs, respective wd is removed from self.WDs,
            # but some other events (like IN_DELETE_SELF) get to this point, without having its reference in self.WDs.
            subscribed_path, _ = self.WDs.get(wd, (None, None))
            to_advertise.add(subscribed_path)
        self.AggrEvents.clear()

        for path in to_advertise:
            if path is None:
                continue
            self.App.PubSub.publish("Library.change!", self, path)

    async def subscribe(self, path):
        if not os.path.isdir(self.BasePath + path):
            return
        if self.FD is None:
            L.warning(
                "Cannot subscribe to changes in the filesystem layer of the library: '{}'".format(
                    self.BasePath
                )
            )
            return
        self._subscribe_recursive(path, path)

    def _subscribe_recursive(self, subscribed_path, path_to_be_listed):
        binary = (self.BasePath + path_to_be_listed).encode()
        wd = inotify_add_watch(self.FD, binary, IN_ALL_EVENTS)
        if wd == -1:
            L.error("Error in inotify_add_watch")
            return
        self.WDs[wd] = (subscribed_path, path_to_be_listed)

        try:
            items = self._list(path_to_be_listed)
        except KeyError:
            # subscribing to non-existing directory is silent
            return

        for item in items:
            if item.type == "dir":
                self._subscribe_recursive(subscribed_path, item.name)

    async def finalize(self, app):
        if self.FD is not None:
            self.App.Loop.remove_reader(self.FD)
            os.close(self.FD)



from .connection import GoogleDriveConnection
from .abcsink import GoogleDriveABCSink
from .blocksink import GoogleDriveBlockSink


__all__ = (
    "GoogleDriveConnection",
    "GoogleDriveABCSink",
    "GoogleDriveBlockSink",
)



import logging

from .analyzer import Analyzer
from ..matrix.geomatrix import GeoMatrix, PersistentGeoMatrix


L = logging.getLogger(__name__)


class GeoAnalyzer(Analyzer):
    """
    This is the analyzer for events with geographical points dimension.

    `GeoAnalyzer` operates over the `GeoMatrix` object.
    `matrix_id` is an id of `GeoMatrix` object defined alternatively.

    **Config Defaults**

    max_lat : 71.26
    min_lat : 23.33
    min_lon : -10.10
    max_lon : 40.6

    """

    ConfigDefaults = {
        "max_lat": 71.26,  # Europe endpoints
        "min_lat": 23.33,
        "min_lon": -10.10,
        "max_lon": 40.6,
    }

    def __init__(
        self,
        app,
        pipeline,
        matrix_id=None,
        dtype="float_",
        analyze_on_clock=False,
        bbox=None,
        resolution=5,
        persistent=False,
        id=None,
        config=None,
    ):
        """
        Description:

        **Parameters**

        app : Application
                Name of the Application.

        pipeline : Pipeline
                Name of the Pipeline.

        matrix_id : str, default = None

        dtype : ??, default = "float_"

        analyze_on_clock : bool, default = False

        bbox : ??, default = None

        resolution : int, default = 5

        persistent : bool, default = False

        id : str, default = None

        config = JSON, default = None
                configuration file with additional information

        """
        super().__init__(
            app, pipeline, analyze_on_clock=analyze_on_clock, id=id, config=config
        )
        svc = app.get_service("bspump.PumpService")
        if matrix_id is None:
            g_id = self.Id + "Matrix"
            if bbox is None:
                bbox = {
                    "min_lat": float(self.ConfigDefaults["min_lat"]),
                    "max_lat": float(self.ConfigDefaults["max_lat"]),
                    "min_lon": float(self.ConfigDefaults["min_lon"]),
                    "max_lon": float(self.ConfigDefaults["max_lon"]),
                }
            if persistent:
                self.GeoMatrix = PersistentGeoMatrix(
                    app, dtype, bbox=bbox, resolution=resolution, id=g_id, config=config
                )
            else:
                self.GeoMatrix = GeoMatrix(
                    app, dtype, bbox=bbox, resolution=resolution, id=g_id, config=config
                )
            svc.add_matrix(self.GeoMatrix)
        else:
            self.GeoMatrix = svc.locate_matrix(matrix_id)



import numpy as np
import logging

from .timewindowanalyzer import TimeWindowAnalyzer

###

L = logging.getLogger(__name__)

###


class ThresholdAnalyzer(TimeWindowAnalyzer):
    """

            Threshold Analyzer is based on TimeWindowAnalyzer and detects, if any
            monitored value exceeded or subceeded the preconfigured bounds.
            This analyzer can be used also only through configuration.

            predicate method - Check whether event contains related attributes

            evaluate method - Take event attributes and sorts them into the matrix

            analyze method - Check whether any value in the matrix is over the preconfigured bounds and if so, check the
                                            occurrance of the symptom in the array and then possibly call the alarm.
                                            To analyze, whether the values are out of bounds, the 'np.where()' method is used. It pass the
                                            position of values out of bounds within the matrix to the alarm method.
                                            x = row position, y = column position
                                            To detect whether the symptom occurrance of the 'values out of bounds' is higher or equal, than number
                                            set in the configuration, an aggregation is used. Result of aggregation is the number of occurrences
                                            and indices of those values. It is then passed to the alarm method together with x and y arrays.

            alarm method - Configurable, takes arguments.

    ...

            Threshold settings:
                    exceedance >>> if 'lower_bound' is set to '-inf' and upper_bound is not set to 'inf', then alarm is called when
                                                    any value in matrix exceed upper_bound

                    subceedance >>> if lower_bound is not set to '-inf' and upper_bound is set to 'inf', then alarm is called when
                                                    any value in matrix subceed lower_bound

                    range >>> if lower_bound is not set to '-inf' and upper_bound is not set to 'inf', then alarm is called when
                                            any value in matrix is out of preconfigured range

    """

    ConfigDefaults = {
        "event_attribute": "",  # Name of the attribute, e.g. server name
        "event_value": "",  # Name of the attribute values to load into the matrix and to be used for
        # analyzing, e.g. server overload values
        "lower_bound": "-inf",  # Lower bound of the threshold
        "upper_bound": "inf",  # Upper bound of the threshold
        "anomaly_occurrence": 1,  # Number of occurrences of anomaly in array
        "analyze_period": 300,  # Launch period of analyze method
    }

    def __init__(self, app, pipeline, id=None, config=None):
        """
        Description:

        **Parameters**

        app : Application
                Name of the Application

        pipeline : Pipeline
                Name of the Pipeline

        id : str, default = None

        config : JSON, default = None
                configuration file with additional information.

        """

        super().__init__(
            app,
            pipeline,
            analyze_on_clock=True,
            clock_driven=True,
            id=id,
            config=config,
        )

        self.EventAttribute = self.Config["event_attribute"]
        self.EventValue = self.Config["event_value"]
        self.Lower = float(self.Config["lower_bound"])
        self.Upper = float(self.Config["upper_bound"])
        self.SymptomOccurrence = int(self.Config["symptom_occurrence"])

        self.WarmingUpLimit = int()

    def predicate(self, context, event):
        """
        Description:

        **Parameters**

        context :

        event : any data type
                Information with timestamp.

        :return: True
        """
        if self.EventAttribute not in event:
            return False

        if "@timestamp" not in event:
            return False

        return True

    def evaluate(self, context, event):
        """
        Description:

        **Parameters**

        context :

        event : any data type
                Information with timestamp.

        """
        attribute = event[self.EventAttribute]
        time_stamp = event["@timestamp"]

        # Getting the row indices
        row = self.TimeWindow.get_row_index(attribute)
        if row is None:
            row = self.TimeWindow.add_row(attribute)

        # Find the column in timewindow matrix to fit in
        column = self.TimeWindow.get_column(time_stamp)
        if column is None:
            return

        # Fill matrix with the values
        self.TimeWindow.Array[row, column] = event[self.EventValue]

    def analyze(self):
        """
        Description:

        """
        # Checking an empty array
        if self.TimeWindow.Array.shape[0] == 0:
            return

        data = self.TimeWindow.Array
        # Warming up the matrix to avoid procedures on not fully filled matrix and resizing warming_up dimension
        # to avoid ValueError when checking the conditions of exceedance/subceedance/range.
        self.WarmingUpLimit = self.TimeWindow.Columns - 1
        warming_up = np.resize(
            self.TimeWindow.WarmingUpCount <= self.WarmingUpLimit, data.shape
        )

        # Exceedance
        if self.Lower == float("-inf") and self.Upper != float("inf"):
            x, y = np.where((data > self.Upper) & warming_up)

        # Subceedance
        elif self.Lower != float("-inf") and self.Upper == float("inf"):
            x, y = np.where((data < self.Lower) & warming_up)

        # Range
        elif self.Lower != float("-inf") and self.Upper != float("inf"):
            x, y = np.where(((data < self.Lower) | (data > self.Upper)) & warming_up)

        # No boundaries set
        else:
            raise ValueError("Boundaries of the threshold has not been set!")

        # Symptom occurrence detection
        try:
            # Setting previous value. When it is empty, further computation wont proceed.
            previous = x[0]
        except IndexError:
            return
        # Setting initial value to count
        count = 1
        for i in range(1, len(x)):
            # Setting actual value
            actual = x[i]
            # Checking, if previous value equals actual
            if actual == previous:
                count += 1
            else:
                # Checking whether count >= occurrence

                if count >= self.SymptomOccurrence:
                    self.alarm(x, y, count, i - 1)
                count = 1
                previous = actual
            # Checking if count >= occurrence and if the loop is at its end
            if (count >= self.SymptomOccurrence) and (i + 1 == len(x)):
                self.alarm(x, y, count, i)

    def alarm(self, *args):
        """
        Description:

        """

        pass



from .analyzer import Analyzer
from .timewindowanalyzer import TimeWindowAnalyzer
from .timedriftanalyzer import TimeDriftAnalyzer
from .sessionanalyzer import SessionAnalyzer
from .geoanalyzer import GeoAnalyzer
from .latch import LatchAnalyzer
from .analyzingsource import AnalyzingSource
from .threshold import ThresholdAnalyzer


__all__ = (
    "Analyzer",
    "TimeWindowAnalyzer",
    "TimeDriftAnalyzer",
    "SessionAnalyzer",
    "GeoAnalyzer",
    "LatchAnalyzer",
    "AnalyzingSource",
    "ThresholdAnalyzer",
)



import logging
import asyncio
import typing

from bspump.asab import Service

#

L = logging.getLogger(__name__)

#


class TaskService(Service):
    """
    Task service is for managed execution of fire-and-forget, one-off, background tasks.
    Task can be a coroutine, `asyncio.Future` or `asyncio.Task` and it is executed in the main event loop.

    The result of the task is collected (and discarded) automatically.
    When the task raises Exception, it will be printed to the log.
    """

    def __init__(self, app, service_name="asab.TaskService"):
        super().__init__(app, service_name)

        self.NewTasks = asyncio.Queue()
        self.PendingTasks = set()
        self.Main = None

    async def initialize(self, app):
        self.start()

    def start(self):
        assert self.Main is None
        self.Main = asyncio.ensure_future(self.main())
        self.Main.add_done_callback(self._main_task_exited)

    async def finalize(self, app):
        if self.Main is not None:
            task = self.Main
            self.Main = None

            task.cancel()
            try:
                await task
            except asyncio.CancelledError:
                pass

        for task in list(self.PendingTasks):
            task.cancel()
            try:
                await task
                self.PendingTasks.remove(task)
            except asyncio.CancelledError:
                self.PendingTasks.remove(task)

        total_tasks = len(self.PendingTasks) + self.NewTasks.qsize()
        if total_tasks > 0:
            L.warning(
                "{}+{} pending and incomplete tasks".format(
                    len(self.PendingTasks), self.NewTasks.qsize()
                )
            )

    def _main_task_exited(self, ctx):
        if self.Main is None:
            return
        try:
            self.Main.result()
        except asyncio.CancelledError:
            pass

        self.Main = None
        L.warning("Main task exited unexpectedly, restarting ...")
        self.start()

    def schedule(self, *tasks):
        """
        Schedule a task (or tasks) for immediate fire-and-forget execution (e.g. compressing files).

        Examples:

        ```python
        app.TaskService.schedule(self._start())
        ```
        """
        for task in tasks:
            self.NewTasks.put_nowait(task)

    def run_forever(self, *async_functions):
        """
        Schedule an async function (or functions) for immediate fire-and-forget execution.
        The function is expected to run forever.
        If function exits, the error is logged and the function is restarted.
        Function is called without any argument.

        Examples:

        ```python
        class MyClass(object):
                def __init__(self, app):
                        ...
                        app.TaskService.run_forever(self.my_forever_method)

                async def my_forever_method(self):
                        while True:
                                await ...
        ```
        """
        for async_fn in async_functions:
            self.NewTasks.put_nowait(forever(async_fn))

    async def main(self):
        while True:
            while self.NewTasks.qsize() > 0:
                task = self.NewTasks.get_nowait()
                if isinstance(task, typing.Coroutine):
                    task = asyncio.create_task(task)
                self.PendingTasks.add(task)

            if len(self.PendingTasks) == 0:
                # Block until a new task is scheduled
                task = await self.NewTasks.get()
                if isinstance(task, typing.Coroutine):
                    task = asyncio.create_task(task)
                self.PendingTasks.add(task)

            else:
                done, self.PendingTasks = await asyncio.wait(
                    self.PendingTasks, timeout=1.0
                )
                for task in done:
                    await task
                    self.App.PubSub.publish("TaskService.task_done!", task)


async def forever(async_fn):
    while True:
        try:
            await async_fn()
        except asyncio.CancelledError:
            break



import base64
import binascii
import datetime
import functools
import inspect
import json
import logging
import os.path
import typing
import time
import enum

import aiohttp
import aiohttp.web
import aiohttp.client_exceptions

from bspump.asab import Config, Service
import bspump.asab.exceptions as exceptions
import bspump.asab.utils as utils

try:
    import jwcrypto.jwk
    import jwcrypto.jwt
    import jwcrypto.jws
except ModuleNotFoundError:
    jwcrypto = None

#

L = logging.getLogger(__name__)

#

# Used for mock authorization
MOCK_USERINFO_DEFAULT = {
    # Token issuer
    "iss": "auth.test.loc",
    # Token issued at (timestamp)
    "iat": int(time.time()),
    # Token expires at (timestamp)
    "exp": int(time.time()) + 5 * 365 * 24 * 3600,
    # Authorized party
    "azp": "my-asab-app",
    # Audience
    "aud": "my-asab-app",
    # Subject (Unique user ID)
    "sub": "abc:xyz:799b53e0",
    # Subject's preferred username
    "preferred_username": "little-capybara",
    # Subject's email
    "email": "capybara1999@example.com",
    # Authorized tenants and resources
    "resources": {
        # Globally granted resources
        "*": [
            "authz:superuser",
        ],
        # Resources granted within the tenant "default"
        "default": [
            "authz:superuser",
            "some-test-data:access",
        ],
        # Resources granted within the tenant "test-tenant"
        "test-tenant": [
            "authz:superuser",
            "cake:eat",
        ],
    },
    # Subject's assigned (not authorized!) tenants
    "tenants": ["default", "test-tenant", "another-tenant"],
}

SUPERUSER_RESOURCE = "authz:superuser"


class AuthMode(enum.Enum):
    ENABLED = enum.auto()
    DISABLED = enum.auto()
    MOCK = enum.auto()


Config.add_defaults(
    {
        "auth": {
            # URL location containing the authorization server's public JWK keys
            # (often found at "/.well-known/jwks.json")
            "public_keys_url": "",
            # Whether the app is tenant-aware
            "multitenancy": "yes",
            # The "enabled" option switches authentication and authorization
            # on, off or activates mock mode. The default value is True (on).
            # In MOCK MODE
            # - no authorization server is needed,
            # - all incoming requests are mock-authorized with pre-defined user info,
            # - custom mock user info can supplied in a JSON file.
            # "enabled": "yes",
            "mock_user_info_path": "/conf/mock-userinfo.json",
        }
    }
)


class AuthService(Service):
    """
    Provides authentication and authorization of incoming requests.
    """

    _PUBLIC_KEYS_URL_DEFAULT = "http://localhost:3081/.well-known/jwks.json"

    def __init__(self, app, service_name="asab.AuthzService"):
        super().__init__(app, service_name)
        self.MultitenancyEnabled = Config.getboolean("auth", "multitenancy")
        self.PublicKeysUrl = Config.get("auth", "public_keys_url")

        enabled = Config.get("auth", "enabled", fallback=True)
        if enabled == "mock":
            self.Mode = AuthMode.MOCK
        elif utils.string_to_boolean(enabled):
            self.Mode = AuthMode.ENABLED
        else:
            self.Mode = AuthMode.DISABLED

        if self.Mode == AuthMode.DISABLED:
            pass
        elif self.Mode == AuthMode.MOCK:
            self.MockUserInfo = self._prepare_mock_user_info()
        elif jwcrypto is None:
            raise ModuleNotFoundError(
                "You are trying to use asab.web.auth module without 'jwcrypto' installed. "
                "Please run 'pip install jwcrypto' "
                "or install asab with 'authz' optional dependency."
            )
        elif len(self.PublicKeysUrl) == 0:
            self.PublicKeysUrl = self._PUBLIC_KEYS_URL_DEFAULT
            L.warning(
                "No 'public_keys_url' provided in [auth] config section. "
                "Defaulting to {!r}.".format(self._PUBLIC_KEYS_URL_DEFAULT)
            )

        self.AuthServerPublicKey = None  # TODO: Support multiple public keys
        # Limit the frequency of auth server requests to save network traffic
        self.AuthServerCheckCooldown = datetime.timedelta(minutes=5)
        self.AuthServerLastSuccessfulCheck = None

    def _prepare_mock_user_info(self):
        # Load custom user info
        mock_user_info_path = Config.get("auth", "mock_user_info_path")
        if os.path.isfile(mock_user_info_path):
            with open(mock_user_info_path, "rb") as fp:
                user_info = json.load(fp)
        else:
            user_info = MOCK_USERINFO_DEFAULT
        # Validate user info
        resources = user_info.get("resources", {})
        if not isinstance(resources, dict) or not all(
            map(
                lambda kv: isinstance(kv[0], str) and isinstance(kv[1], list),
                resources.items(),
            )
        ):
            raise ValueError(
                "User info 'resources' must be an object with string keys and array values."
            )
        L.warning(
            "AuthService is running in MOCK MODE. All web requests will be authorized with mock user info, which "
            "currently grants access to the following tenants: {}. To customize mock mode authorization (add or "
            "remove tenants and resources, change username etc.), provide your own user info in {!r}.".format(
                list(t for t in user_info.get("resources", {}).keys() if t != "*"),
                mock_user_info_path,
            )
        )
        return user_info

    async def initialize(self, app):
        if self.Mode == AuthMode.ENABLED:
            await self._fetch_public_keys_if_needed()

    def is_enabled(self) -> bool:
        """
        Check if the AuthService is enabled. Mock mode counts as enabled too.
        """
        return self.Mode in {AuthMode.ENABLED, AuthMode.MOCK}

    def install(self, web_container):
        """
        Apply authorization to all web handlers in a web container, according to their arguments and path parameters.

        :param web_container: Web container to be protected by authorization.
        :type web_container: asab.web.WebContainer
        """
        # TODO: Call this automatically if there is only one container
        web_container.WebApp.on_startup.append(self._wrap_handlers)

    def is_ready(self):
        """
        Check if the service is ready to authorize requests.
        """
        if self.Mode == AuthMode.DISABLED:
            return True
        if self.Mode == AuthMode.MOCK:
            return True
        if self.AuthServerPublicKey is None:
            return False
        return True

    async def get_userinfo_from_id_token(self, bearer_token):
        """
        Parse the bearer ID token and extract user info.
        """
        if not self.is_ready():
            # Try to load the public keys again
            if self.AuthServerPublicKey is None:
                await self._fetch_public_keys_if_needed()
            if not self.is_ready():
                L.error(
                    "Cannot authenticate request: Failed to load authorization server's public keys."
                )
                raise aiohttp.web.HTTPUnauthorized()

        try:
            return _get_id_token_claims(bearer_token, self.AuthServerPublicKey)
        except jwcrypto.jws.InvalidJWSSignature:
            # Authz server keys may have changed. Try to reload them.
            L.warning("Invalid ID token signature.")
            await self._fetch_public_keys_if_needed()

        try:
            return _get_id_token_claims(bearer_token, self.AuthServerPublicKey)
        except jwcrypto.jws.InvalidJWSSignature:
            L.error("Cannot authenticate request: Invalid ID token signature.")
            raise exceptions.NotAuthenticatedError()

    def has_superuser_access(self, authorized_resources: typing.Iterable) -> bool:
        """
        Check if the superuser resource is present in the authorized resource list.
        """
        if self.Mode == AuthMode.DISABLED:
            return True
        return SUPERUSER_RESOURCE in authorized_resources

    def has_resource_access(
        self, authorized_resources: typing.Iterable, required_resources: typing.Iterable
    ) -> bool:
        """
        Check if the requested resources or the superuser resource are present in the authorized resource list.
        """
        if self.Mode == AuthMode.DISABLED:
            return True
        if self.has_superuser_access(authorized_resources):
            return True
        for resource in required_resources:
            if resource not in authorized_resources:
                return False
        return True

    async def _fetch_public_keys_if_needed(self, *args, **kwargs):
        """
        Check if public keys have been fetched from the authorization server and fetch them if not yet.
        """
        now = datetime.datetime.now(datetime.timezone.utc)
        if (
            self.AuthServerLastSuccessfulCheck is not None
            and now < self.AuthServerLastSuccessfulCheck + self.AuthServerCheckCooldown
        ):
            # Public keys have been fetched recently
            return

        async with aiohttp.ClientSession() as session:
            try:
                async with session.get(self.PublicKeysUrl) as response:
                    if response.status != 200:
                        L.error(
                            "HTTP error while loading public keys.",
                            struct_data={
                                "status": response.status,
                                "url": self.PublicKeysUrl,
                                "text": await response.text(),
                            },
                        )
                        return
                    try:
                        data = await response.json()
                    except json.JSONDecodeError:
                        L.error(
                            "JSON decoding error while loading public keys.",
                            struct_data={
                                "url": self.PublicKeysUrl,
                                "data": data,
                            },
                        )
                        return
                    try:
                        key_data = data["keys"].pop()
                    except (IndexError, KeyError):
                        L.error(
                            "Error while loading public keys: No public keys in server response.",
                            struct_data={
                                "url": self.PublicKeysUrl,
                                "data": data,
                            },
                        )
                        return
                    try:
                        public_key = jwcrypto.jwk.JWK(**key_data)
                    except Exception as e:
                        L.error(
                            "JWK decoding error while loading public keys: {}.".format(
                                e
                            ),
                            struct_data={
                                "url": self.PublicKeysUrl,
                                "data": data,
                            },
                        )
                        return
            except aiohttp.client_exceptions.ClientConnectorError as e:
                L.error(
                    "Connection error while loading public keys: {}".format(e),
                    struct_data={
                        "url": self.PublicKeysUrl,
                    },
                )
                return

        self.AuthServerPublicKey = public_key
        self.AuthServerLastSuccessfulCheck = datetime.datetime.now(
            datetime.timezone.utc
        )
        L.debug("Public key loaded.", struct_data={"url": self.PublicKeysUrl})

    def _authenticate_request(self, handler):
        """
        Authenticate the request by the JWT ID token in the Authorization header.
        Extract the token claims into request attributes so that they can be used for authorization checks.
        """

        @functools.wraps(handler)
        async def wrapper(*args, **kwargs):
            request = args[-1]
            if self.Mode == AuthMode.DISABLED:
                user_info = None
            elif self.Mode == AuthMode.MOCK:
                user_info = self.MockUserInfo
            else:
                # Extract user info from the request Authorization header
                bearer_token = _get_bearer_token(request)
                user_info = await self.get_userinfo_from_id_token(bearer_token)

            # Add userinfo, tenants and global resources to the request
            if self.Mode != AuthMode.DISABLED:
                assert user_info is not None
                request._UserInfo = user_info
                resource_dict = request._UserInfo["resources"]
                request._Resources = frozenset(resource_dict.get("*", []))
                request._Tenants = frozenset(
                    t for t in resource_dict.keys() if t != "*"
                )
            else:
                request._UserInfo = None
                request._Resources = None
                request._Tenants = None

            # Add access control methods to the request
            def has_resource_access(*required_resources: list) -> bool:
                return self.has_resource_access(request._Resources, required_resources)

            request.has_resource_access = has_resource_access

            def has_superuser_access() -> bool:
                return self.has_superuser_access(request._Resources)

            request.has_superuser_access = has_superuser_access

            return await handler(*args, **kwargs)

        return wrapper

    async def _wrap_handlers(self, aiohttp_app):
        """
        Inspect all registered handlers and wrap them in decorators according to their parameters.
        """
        for route in aiohttp_app.router.routes():
            # Skip non-coroutines
            if not inspect.iscoroutinefunction(route.handler):
                continue

            # Skip auth for HEAD requests
            if route.method == "HEAD":
                continue

            try:
                self._wrap_handler(route)
            except Exception as e:
                raise Exception(
                    "Failed to initialize auth for handler {!r}.".format(
                        route.handler.__qualname__
                    )
                ) from e

    def _wrap_handler(self, route):
        """
        Inspect handler and apply suitable auth wrappers.
        """
        # Check if tenant is in route path
        route_info = route.get_info()
        tenant_in_path = (
            "formatter" in route_info and "{tenant}" in route_info["formatter"]
        )

        # Extract the actual handler method for signature checks
        handler_method = route.handler
        while hasattr(handler_method, "__wrapped__"):
            # While loop unwraps handlers wrapped in multiple decorators.
            # NOTE: This requires all the decorators to use @functools.wraps().
            handler_method = handler_method.__wrapped__

        if hasattr(handler_method, "__func__"):
            handler_method = handler_method.__func__

        if hasattr(handler_method, "NoAuth"):
            return
        argspec = inspect.getfullargspec(handler_method)
        args = set(argspec.kwonlyargs).union(argspec.args)

        # Extract the whole handler for wrapping
        handler = route.handler

        # Apply the decorators in reverse order (the last applied wrapper affects the request first)
        if "resources" in args:
            handler = _add_resources(handler)
        if "user_info" in args:
            handler = _add_user_info(handler)
        if "tenant" in args:
            if tenant_in_path:
                handler = self._add_tenant_from_path(handler)
            elif self.MultitenancyEnabled or self.Mode == AuthMode.MOCK:
                handler = self._add_tenant_from_query(handler)
            else:
                handler = self._add_tenant_none(handler)

        handler = self._authenticate_request(handler)
        route._handler = handler

    def _authorize_tenant_request(self, request, tenant):
        """
        Check access to requested tenant and add tenant resources to the request
        """
        # Check if tenant access is authorized
        if tenant not in request._Tenants:
            L.warning(
                "Tenant not authorized.",
                struct_data={"tenant": tenant, "sub": request._UserInfo.get("sub")},
            )
            raise exceptions.AccessDeniedError()

        # Extend globally granted resources with tenant-granted resources
        request._Resources = frozenset(
            request._Resources.union(request._UserInfo["resources"].get(tenant, []))
        )

    def _add_tenant_from_path(self, handler):
        """
        Extract tenant from request path and authorize it
        """

        @functools.wraps(handler)
        async def wrapper(*args, **kwargs):
            request = args[-1]
            tenant = request.match_info["tenant"]
            if self.Mode != AuthMode.DISABLED:
                self._authorize_tenant_request(request, tenant)
            return await handler(*args, tenant=tenant, **kwargs)

        return wrapper

    def _add_tenant_from_query(self, handler):
        """
        Extract tenant from request query and authorize it
        """

        @functools.wraps(handler)
        async def wrapper(*args, **kwargs):
            request = args[-1]
            if "tenant" not in request.query:
                if self.Mode != AuthMode.MOCK:
                    L.error("Request is missing 'tenant' query parameter.")
                    raise aiohttp.web.HTTPBadRequest()
                tenant = None
            else:
                tenant = request.query["tenant"]
                if self.Mode != AuthMode.DISABLED:
                    self._authorize_tenant_request(request, tenant)
            return await handler(*args, tenant=tenant, **kwargs)

        return wrapper

    def _add_tenant_none(self, handler):
        """
        Add tenant=None to the handler arguments
        """

        @functools.wraps(handler)
        async def wrapper(*args, **kwargs):
            return await handler(*args, tenant=None, **kwargs)

        return wrapper


def _get_id_token_claims(bearer_token: str, auth_server_public_key):
    """
    Parse and validate JWT ID token and extract the claims (user info)
    """
    assert jwcrypto is not None
    try:
        token = jwcrypto.jwt.JWT(jwt=bearer_token, key=auth_server_public_key)
    except jwcrypto.jwt.JWTExpired:
        L.warning("ID token expired.")
        raise exceptions.NotAuthenticatedError()
    except jwcrypto.jws.InvalidJWSSignature as e:
        raise e
    except ValueError as e:
        L.error(
            "Failed to parse JWT ID token ({}). Please check if the Authorization header contains ID token.".format(
                e
            )
        )
        raise aiohttp.web.HTTPBadRequest()
    except Exception:
        L.exception(
            "Failed to parse JWT ID token. Please check if the Authorization header contains ID token."
        )
        raise aiohttp.web.HTTPBadRequest()

    try:
        token_claims = json.loads(token.claims)
    except Exception:
        L.exception("Failed to parse JWT token claims.")
        raise aiohttp.web.HTTPBadRequest()

    return token_claims


def _get_id_token_claims_without_verification(bearer_token: str):
    """
    Parse JWT ID token without validation and extract the claims (user info)
    """
    try:
        header, payload, signature = bearer_token.split(".")
    except IndexError:
        L.warning("Cannot parse ID token: Wrong number of '.'.")
        raise aiohttp.web.HTTPBadRequest()

    try:
        claims = json.loads(base64.b64decode(payload.encode("utf-8")))
    except binascii.Error:
        L.warning("Cannot parse ID token: Payload is not base 64.")
        raise aiohttp.web.HTTPBadRequest()
    except json.JSONDecodeError:
        L.warning("Cannot parse ID token: Payload cannot be parsed as JSON.")
        raise aiohttp.web.HTTPBadRequest()

    return claims


def _get_bearer_token(request):
    """
    Validate the Authorizetion header and extract the Bearer token value
    """
    authorization_header = request.headers.get(aiohttp.hdrs.AUTHORIZATION)
    if authorization_header is None:
        L.warning("No Authorization header.")
        raise aiohttp.web.HTTPUnauthorized()
    try:
        auth_type, token_value = authorization_header.split(" ", 1)
    except ValueError:
        L.warning("Cannot parse Authorization header.")
        raise aiohttp.web.HTTPBadRequest()
    if auth_type != "Bearer":
        L.warning("Unsupported Authorization header type: {!r}".format(auth_type))
        raise aiohttp.web.HTTPUnauthorized()
    return token_value


def _add_user_info(handler):
    """
    Add user info to the handler arguments
    """

    @functools.wraps(handler)
    async def wrapper(*args, **kwargs):
        request = args[-1]
        return await handler(*args, user_info=request._UserInfo, **kwargs)

    return wrapper


def _add_resources(handler):
    """
    Add resources to the handler arguments
    """

    @functools.wraps(handler)
    async def wrapper(*args, **kwargs):
        request = args[-1]
        return await handler(*args, resources=request._Resources, **kwargs)

    return wrapper



import logging
import os
import struct

import aiohttp
import asyncio

from bspump.asab import Config

from bspump.abc.lookupprovider import LookupBatchProviderABC

###

L = logging.getLogger(__name__)

###


class HTTPBatchLookupProvider(LookupBatchProviderABC):
    """
    Fetches lookup data from given URL over HTTP. This lookupprovider embeds loading and caching functions of the
    original bspump.Lookup in "slave" mode.
    """

    ConfigDefaults = {"master_timeout": 30, "use_cache": "yes", "cache_dir": ""}

    def __init__(self, lookup, url, id=None, config=None):
        super().__init__(lookup, url, id, config=config)
        self.Timeout = self.Config.get("master_timeout")
        self.UseCache = self.Config.getboolean("use_cache")
        if self.Timeout is not None:
            self.Timeout = float(self.Timeout)
        self.CachePath = None
        if self.UseCache is True:
            cache_path = self.Config.get("cache_dir", "").strip()
            if (
                len(cache_path) == 0
                and "general" in Config
                and "var_dir" in Config["general"]
            ):
                cache_path = os.path.abspath(Config["general"]["var_dir"])
                self.CachePath = os.path.join(
                    cache_path, "lookup_{}.cache".format(self.Id)
                )
            else:
                self.UseCache = False
                L.warning("No cache path specified. Cache disabled.")

    async def load(self):
        headers = {}
        if self.ETag is not None:
            headers["ETag"] = self.ETag

        async with aiohttp.ClientSession() as session:
            try:
                response = await session.get(
                    self.URL,
                    headers=headers,
                    timeout=float(self.Config["master_timeout"]),
                )
            except aiohttp.ClientConnectorError as e:
                L.warning(
                    "{}: Failed to contact lookup master at '{}': {}".format(
                        self.Id, self.URL, e
                    )
                )
                return self.load_from_cache()
            except asyncio.TimeoutError as e:
                L.warning(
                    "{}: Failed to contact lookup master at '{}' (timeout): {}".format(
                        self.Id, self.URL, e
                    )
                )
                return self.load_from_cache()

            if response.status == 304:
                L.info("Lookup '{}' is up to date at {}.".format(self.Id, self.URL))
                return False

            if response.status == 404:
                L.warning(
                    "{}: lookup was not found at the lookup provider {}".format(
                        self.Id, self.URL
                    )
                )
                return self.load_from_cache()

            if response.status == 501:
                L.warning(
                    "{}: master '{}' does not support serialization.".format(
                        self.Id, self.URL
                    )
                )
                return False

            if response.status != 200:
                L.warning(
                    "{}: Failed to get lookup from master {}".format(self.Id, self.URL)
                )
                return self.load_from_cache()
            data = await response.read()
            self.ETag = response.headers.get("ETag")
            if self.CachePath is not None:
                self.save_to_cache(data)
        return data

    def load_from_cache(self):
        """
        Load the lookup data (bytes) from cache.
        """
        if self.UseCache is False:
            return False
        # Load the ETag from cached file, if have one
        if not os.path.isfile(self.CachePath):
            L.warning("Cache '{}': not a file".format(self.CachePath))
            return False

        if not os.access(self.CachePath, os.R_OK):
            L.warning("Cannot read cache from '{}'".format(self.CachePath))
            return False

        try:
            with open(self.CachePath, "rb") as f:
                (tlen,) = struct.unpack(r"<L", f.read(struct.calcsize(r"<L")))
                etag_b = f.read(tlen)
                self.ETag = etag_b.decode("utf-8")
                f.read(1)
                data = f.read()
            return data
        except Exception as e:
            L.warning(
                "Failed to read content of lookup cache '{}' from '{}': {}".format(
                    self.Id, self.CachePath, e
                )
            )
            os.unlink(self.CachePath)
        return False

    def save_to_cache(self, data):
        if self.UseCache is False:
            return
        dirname = os.path.dirname(self.CachePath)
        if not os.path.isdir(dirname):
            os.makedirs(dirname)

        with open(self.CachePath, "wb") as fo:
            # Write E-Tag and '\n'
            etag_b = self.ETag.encode("utf-8")
            fo.write(struct.pack(r"<L", len(etag_b)) + etag_b + b"\n")

            # Write Data
            fo.write(data)



import time
import logging

import numpy as np

from bspump.asab import Timer

import os

from .namedmatrix import NamedMatrix, PersistentNamedMatrix
from .utils.timeconfig import TimeConfig, PersistentTimeConfig
from .utils.warmingupcount import WarmingUpCount, PersistentWarmingUpCount

###

L = logging.getLogger(__name__)

###


class TimeWindowMatrix(NamedMatrix):
    """
            Container, specific for `TimeWindowAnalyzer`.

            The main specific attributes are:
            `Start` is the starting timestamp of the first column of the matrix;
            `End` is the ending timestamp of the last column;
            `Resolution` is the width of the column in seconds.

    .. code-block:: python

            --> Columns (time dimension), column "width" = resolution
            +---+---+---+---+---+---+
            |   |   |   |   |   |   |
            +---+---+---+---+---+---+
            |   |   |   |   |   |   |
            +---+---+---+---+---+---+
            |   |   |   |   |   |   |
            +---+---+---+---+---+---+
            |   |   |   |   |   |   |
            +---+---+---+---+---+---+
            |   |   |   |   |   |   |
            +---+---+---+---+---+---+
            ^                       ^
            End (past)   <          Start (== now)

    """

    def __init__(
        self,
        app,
        dtype="float_",
        start_time=None,
        resolution=60,
        columns=15,
        clock_driven=False,
        id=None,
        config=None,
    ):
        self.Columns = columns
        if start_time is None:
            start_time = time.time()

        start = (1 + (start_time // resolution)) * resolution
        self.Start = start
        self.Resolution = resolution

        super().__init__(app, dtype=dtype, id=id, config=config)

        if clock_driven:
            advance_period = resolution / 4
            self.Timer = Timer(app, self.on_clock_tick, autorestart=True)
            self.Timer.start(advance_period)
        else:
            self.Timer = None

        self.ClockDriven = clock_driven
        metrics_service = app.get_service("asab.MetricsService")
        self.Counters = metrics_service.create_counter(
            "EarlyLateEventCounter",
            tags={
                "matrix": self.Id,
            },
            init_values={
                "events.early": 0,
                "events.late": 0,
            },
        )

    def build_shape(self, rows=0):
        """
        Override this method to have a control over the shape of the matrix.
        """
        return (
            rows,
            self.Columns,
        )

    def reshape(self, shape):
        return (
            int(shape[0] / self.Columns),
            self.Columns,
        )

    def add_row(self, row_name):
        """
        Adds new row with `row_id` to the matrix and sets `warming_up_count`.
        """

        row_index = super().add_row(row_name)
        if self.Array.shape[0] != len(self.WarmingUpCount):
            self.WarmingUpCount.extend(self.Array.shape[0], self.Array.shape[1])
        else:
            self.WarmingUpCount.assign(row_index, self.Array.shape[1])

        return row_index

    def get_column(self, event_timestamp):
        """
        Returns the right column, where the timestamp fits.
        If if falls earlier or later, returns `None`.
        The timestamp should be provided in seconds.
        """

        if event_timestamp <= self.TimeConfig.get_end():
            self.Counters.add("events.late", 1)
            return None

        if event_timestamp >= self.TimeConfig.get_start():
            self.Counters.add("events.early", 1)
            return None

        column_idx = int(
            (event_timestamp - self.TimeConfig.get_end())
            // self.TimeConfig.get_resolution()
        )

        # These are temporal debug lines
        if column_idx < 0:
            L.exception(
                "The column index {} is less then 0, {} event timestamp, {} start time, {} end time, {} resolution, {} num columns".format(
                    column_idx,
                    event_timestamp,
                    self.TimeConfig.get_start(),
                    self.TimeConfig.get_end(),
                    self.TimeConfig.get_resolution(),
                    self.Array.shape[1],
                )
            )
            raise

        if column_idx >= self.Array.shape[1]:
            L.exception(
                "The column index {} is more then columns number, {} event timestamp, {} start time, {} end time, {} resolution, {} num columns".format(
                    column_idx,
                    event_timestamp,
                    self.TimeConfig.get_start(),
                    self.TimeConfig.get_end(),
                    self.TimeConfig.get_resolution(),
                    self.Array.shape[1],
                )
            )
            raise

        return column_idx

    def advance(self, target_ts):
        """
                Advance time window (add columns) so it covers target `timestamp` (`target_ts`)
                Also, if `target_ts` is in top 75% of the last existing column, add a new column too.

                "target_ts" must always be in seconds

        .. code-block:: python

                --------------------|-----------
                target_ts  ^ >>>    |
                                                        ^
                                                        Start
                ------------------------------

        """
        added = 0
        while True:
            dt = (
                self.TimeConfig.get_start() - target_ts
            ) / self.TimeConfig.get_resolution()
            if dt > 0.25:
                break
            self.add_column()
            added += 1

        return added

    def close_row(self, row_name, clear=True):
        closed = super().close_row(row_name, clear=clear)
        if closed:
            row_index = self.Index.get_row_index(row_name)
            if row_index is not None:
                self.WarmingUpCount.assign(row_index, self.Array.shape[1])

        return closed

    def flush(self):
        """
        The matrix will be recreated without rows from `ClosedRows`.
        """
        closed_indexes, saved_indexes = super().flush()
        self.WarmingUpCount.flush(saved_indexes)
        return closed_indexes, saved_indexes

    async def on_clock_tick(self):
        """
        React on timer's tick and advance the window.
        """
        target_ts = time.time()
        self.advance(target_ts)

    def zeros(self):
        super().zeros()
        self.TimeConfig = TimeConfig(self.Resolution, self.Columns, self.Start)
        self.End = self.TimeConfig.get_end()
        self.WarmingUpCount = WarmingUpCount(self.Array.shape[0])

    def add_column(self):
        """
        Adds new time column to the matrix and deletes the first one, simulating
        the time flow. `Start` and `End` attributes are advanced as well.
        """

        self.TimeConfig.add_start(self.TimeConfig.get_resolution())
        self.TimeConfig.add_end(self.TimeConfig.get_resolution())

        if self.Array.shape[0] == 0:
            return

        column = np.empty(
            (
                self.Array.shape[0],
                1,
            )
            + self.Array.shape[2:],
            dtype=self.Array.dtype,
        )
        column[:] = np.nan
        array = self.Array
        array = np.hstack((array, column))
        array = np.delete(array, 0, axis=1)
        self.Array = array

        open_rows = list(
            set(range(0, self.Array.shape[0])) - self.ClosedRows.get_rows()
        )
        self.WarmingUpCount.decrease(open_rows)
        self.Start = self.TimeConfig.get_start()
        self.End = self.TimeConfig.get_end()


class PersistentTimeWindowMatrix(PersistentNamedMatrix):
    def __init__(
        self,
        app,
        dtype="float_",
        start_time=None,
        resolution=60,
        columns=15,
        clock_driven=False,
        id=None,
        config=None,
    ):
        self.Columns = columns
        if start_time is None:
            start_time = time.time()

        start = (1 + (start_time // resolution)) * resolution
        self.Start = start
        self.Resolution = resolution

        super().__init__(app, dtype=dtype, id=id, config=config)

        if clock_driven:
            advance_period = resolution / 4
            self.Timer = Timer(app, self.on_clock_tick, autorestart=True)
            self.Timer.start(advance_period)
            if self.TimeConfig.get_start() != start:
                self.advance(start)
        else:
            self.Timer = None

        self.ClockDriven = clock_driven
        self.Start = self.TimeConfig.get_start()
        self.End = self.TimeConfig.get_end()
        metrics_service = app.get_service("asab.MetricsService")
        self.Counters = metrics_service.create_counter(
            "EarlyLateEventCounter",
            tags={
                "matrix": self.Id,
            },
            init_values={
                "events.early": 0,
                "events.late": 0,
            },
        )

    def build_shape(self, rows=0):
        """
        Override this method to have a control over the shape of the matrix.
        """
        return (
            rows,
            self.Columns,
        )

    def reshape(self, shape):
        return (
            int(shape[0] / self.Columns),
            self.Columns,
        )

    def add_row(self, row_name):
        """
        Adds new row with `row_id` to the matrix and sets `warming_up_count`.
        """

        row_index = super().add_row(row_name)
        if self.Array.shape[0] != len(self.WarmingUpCount):
            self.WarmingUpCount.extend(self.Array.shape[0], self.Array.shape[1])
        else:
            self.WarmingUpCount.assign(row_index, self.Array.shape[1])

        return row_index

    def get_column(self, event_timestamp):
        """
        Returns the right column, where the timestamp fits.
        If if falls earlier or later, returns `None`.
        The timestamp should be provided in seconds.
        """

        if event_timestamp <= self.TimeConfig.get_end():
            self.Counters.add("events.late", 1)
            return None

        if event_timestamp >= self.TimeConfig.get_start():
            self.Counters.add("events.early", 1)
            return None

        column_idx = int(
            (event_timestamp - self.TimeConfig.get_end())
            // self.TimeConfig.get_resolution()
        )

        # These are temporal debug lines
        if column_idx < 0:
            L.exception(
                "The column index {} is less then 0, {} event timestamp, {} start time, {} end time, {} resolution, {} num columns".format(
                    column_idx,
                    event_timestamp,
                    self.TimeConfig.get_start(),
                    self.TimeConfig.get_end(),
                    self.TimeConfig.get_resolution(),
                    self.Array.shape[1],
                )
            )
            raise

        if column_idx >= self.Array.shape[1]:
            L.exception(
                "The column index {} is more then columns number, {} event timestamp, {} start time, {} end time, {} resolution, {} num columns".format(
                    column_idx,
                    event_timestamp,
                    self.TimeConfig.get_start(),
                    self.TimeConfig.get_end(),
                    self.TimeConfig.get_resolution(),
                    self.Array.shape[1],
                )
            )
            raise

        return column_idx

    def advance(self, target_ts):
        """
                Advance time window (add columns) so it covers target `timestamp` (`target_ts`)
                Also, if `target_ts` is in top 75% of the last existing column, add a new column too.

                "target_ts" must always be in seconds

        .. code-block:: python

                --------------------|-----------
                target_ts  ^ >>>    |
                                                        ^
                                                        Start
                ------------------------------

        """
        added = 0
        while True:
            dt = (
                self.TimeConfig.get_start() - target_ts
            ) / self.TimeConfig.get_resolution()
            if dt > 0.25:
                break
            self.add_column()
            added += 1

        return added

    def close_row(self, row_name, clear=True):
        closed = super().close_row(row_name, clear=clear)
        if closed:
            row_index = self.Index.get_row_index(row_name)
            if row_index is not None:
                self.WarmingUpCount.assign(row_index, self.Array.shape[1])

        return closed

    def flush(self):
        """
        The matrix will be recreated without rows from `ClosedRows`.
        """
        closed_indexes, saved_indexes = super().flush()
        self.WarmingUpCount.flush(saved_indexes)
        return closed_indexes, saved_indexes

    async def on_clock_tick(self):
        """
        React on timer's tick and advance the window.
        """
        target_ts = time.time()
        self.advance(target_ts)

    def zeros(self):
        super().zeros()
        path = os.path.join(self.Path, "time_config.dat")
        self.TimeConfig = PersistentTimeConfig(
            path, self.Resolution, self.Columns, self.Start
        )
        self.End = self.TimeConfig.get_end()
        path = os.path.join(self.Path, "warming_up_count.dat")
        self.WarmingUpCount = PersistentWarmingUpCount(path, self.Array.shape[0])
        if self.TimeConfig.get_start() != self.Start:
            self.advance(self.Start)

    def add_column(self):
        """
        Adds new time column to the matrix and deletes the first one, simulating
        the time flow. `Start` and `End` attributes are advanced as well.
        """

        self.TimeConfig.add_start(self.TimeConfig.get_resolution())
        self.TimeConfig.add_end(self.TimeConfig.get_resolution())

        if self.Array.shape[0] == 0:
            return

        column = np.zeros(
            (
                self.Array.shape[0],
                1,
            )
            + self.Array.shape[2:],
            dtype=self.Array.dtype,
        )
        array = np.zeros(self.Array.shape, dtype=self.DType)
        array[:] = self.Array[:]

        array = np.hstack((array, column))
        array = np.delete(array, 0, axis=1)

        self.Array = np.memmap(
            self.ArrayPath, dtype=self.DType, mode="w+", shape=array.shape
        )
        self.Array[:] = array[:]

        open_rows = list(
            set(range(0, self.Array.shape[0])) - self.ClosedRows.get_rows()
        )
        self.WarmingUpCount.decrease(open_rows)
        self.Start = self.TimeConfig.get_start()
        self.End = self.TimeConfig.get_end()



import asyncio
import logging

import aiohttp.web

from ...abc.source import Source


L = logging.getLogger(__name__)


class WebSocketSource(Source):
    def __init__(self, app, pipeline, id=None, config=None):
        super().__init__(app, pipeline, id=id, config=config)

        self.WebSockets = set()
        app.PubSub.subscribe("Application.exit!", self._on_exit)

    async def main(self):
        pass

    async def handler(self, request):
        ws = aiohttp.web.WebSocketResponse()
        await ws.prepare(request)

        self.WebSockets.add(ws)

        try:
            async for msg in ws:
                if (
                    msg.type == aiohttp.WSMsgType.BINARY
                    or msg.type == aiohttp.WSMsgType.TEXT
                ):
                    await self.Pipeline.process(
                        msg.data, context={"type": msg.type, "request": request}
                    )

                elif msg.type == aiohttp.WSMsgType.ERROR:
                    L.warning(
                        "WebSocket connection closed with exception {}".format(
                            ws.exception()
                        )
                    )

                else:
                    L.warning("WebSocket unknown/invalid message {}".format(msg))

        finally:
            self.WebSockets.remove(ws)

        return ws

    async def _on_exit(self, event_name):
        # Close all existing web socket connections ...
        tasks = [ws.close() for ws in self.WebSockets]
        while len(tasks) > 0:
            done, tasks = await asyncio.wait(tasks)



import abc


class Service(abc.ABC):
    """
    Abstract class for ASAB services.

    Service objects are registered at the service registry `asab.Application.Services`, managed by an application object.

    Examples:

    This is how Service is created and registered:
    ```python
    my_service = MyService(app, "my_service")
    ```

    This is how Service is located and used:

    ```python
    my_service = app.get_service("my_service")
    my_service.service_method()
    ```

    Example of a typical Service class skeleton:

    ```python
    class MyService(asab.Service):
            def __init__(self, app, service_name):
                    super().__init__(app, service_name)
                    ...

            async def initialize(self, app):
                    ...

            async def finalize(self, app):
                    ...

            def service_method(self):
                    ...
    ```
    """

    def __init__(self, app, service_name: str):
        """
        Register the service to `asab.Application.Services` dictionary with the provided `service_name`.

        Args:
                app (asab.Application): Reference to ASAB application.
                service_name: Reference name of the Service.
        """
        self.Name = service_name
        """
		The reference name for the Service.
		"""
        self.App = app
        """
		The reference to the Application object.
		"""
        app._register_service(self)

    # Lifecycle

    async def initialize(self, app):
        """
        This method is called when the Service is initialized.
        It can be overridden by an user.

        Args:
                app (asab.Application): Reference to ASAB application.
        """
        pass

    async def finalize(self, app):
        """
        This method is called when the Service is finalized, e.g., during application `exit-time`.
        It can be overridden by an user.

        Args:
                app (asab.Application): Reference to ASAB application.
        """
        pass



"""
OBSOLETE MODULE, to be deleted after January 2024

Use 'asab.web.auth' instead

"""

from .decorator import required, userinfo_handler
from .middleware import authz_middleware_factory
from .service import AuthzService

__all__ = ("required", "userinfo_handler", "authz_middleware_factory", "AuthzService")



from .connection import ElasticSearchConnection
from .sink import ElasticSearchSink
from .source import ElasticSearchSource, ElasticSearchAggsSource
from .lookup import ElasticSearchLookup

__all__ = [
    "ElasticSearchConnection",
    "ElasticSearchSink",
    "ElasticSearchSource",
    "ElasticSearchAggsSource",
    "ElasticSearchLookup",
]



import json
import typing

import asyncio

from .sink import KafkaSink


class KafkaBatchSink(KafkaSink):
    """
    KafkaBatchSink is a sink processor that forwards the event to
    an Apache Kafka specified by a KafkaConnection object in batches.

    It is a proof of concept sink, that allows faster processing of events in the pipeline,
    but does not guarantee processing of all events in situations when the pump is closed etc.

    There is a work to be done with cooperation with aiokafka, so the send_and_wait method works
    properly and is able to send events in batches.
    """

    ConfigDefaults = {
        "batch_size": 10000,
    }

    def __init__(
        self, app, pipeline, connection, key_serializer=None, id=None, config=None
    ):
        """
        Initializing parameters passed to the BatchSink class.

        **Parameters**

        app : Application
                Name of the `Application <https://asab.readthedocs.io/en/latest/asab/application.html#>`_.

        pipeline : Pipeline
                Name of the Pipeline.

        connection : Connection
                Information needed to creates connection.

        key_serializer :  ,default None

        id :  , default = None

        config : JSON, default = None
                Configuration file with additional information.

        """
        super().__init__(app, pipeline, connection, key_serializer, id, config)
        self._batch_size = int(self.Config["batch_size"])
        self._throttled = False

    def process(self, context, event: typing.Union[dict, str, bytes]):
        """
        Starts the sink process.

        **Parameters**

        context : type?
                Additional information.

        event: typing.Union[dict, str, bytes] : type?

        """
        if type(event) == dict:
            event = json.dumps(event)
            event = event.encode(self.Encoding)
        elif type(event) == str:
            event = event.encode(self.Encoding)
        kafka_topic = context.get("kafka_topic", self.Topic)
        kafka_key = context.get("kafka_key")

        if self._key_serializer is not None and kafka_key is not None:
            kafka_key = self._key_serializer(kafka_key)

        self._output_queue.put_nowait((kafka_topic, event, kafka_key))

        if (
            not self._throttled
            and self._output_queue.qsize() >= self._output_queue_max_size
        ):
            self.Pipeline.throttle(self, True)
            self._throttled = True

    async def _connection(self):
        producer = await self.Connection.create_producer(**self._producer_params)
        try:
            await producer.start()
            while True:
                futures = []
                iterations = 0
                while iterations < self._batch_size:
                    iterations = iterations + 1
                    topic, message, kafka_key = await self._output_queue.get()

                    if topic is None and message is None:
                        continue
                    future = await producer.send(topic, message, key=kafka_key)
                    futures.append(future)

                    if (
                        self._throttled
                        and self._output_queue.qsize() < self._output_queue_max_size
                    ):
                        self.Pipeline.throttle(self, False)
                        self._throttled = False

                if futures:
                    await asyncio.gather(*futures)

        finally:
            await producer.stop()



import logging

import pymysql
import pymysql.cursors
import pymysql.err

import aiomysql.cursors

from ..abc.source import TriggerSource


L = logging.getLogger(__name__)


class MySQLSource(TriggerSource):
    ConfigDefaults = {"query": ""}

    def __init__(self, app, pipeline, connection, id=None, config=None):
        super().__init__(app, pipeline, id=id, config=config)
        self.Loop = app.Loop
        self._connection = pipeline.locate_connection(app, connection)
        self._query = self.Config["query"]
        self.App = app

    async def cycle(self):
        await self._connection.ConnectionEvent.wait()
        try:
            async with self._connection.acquire_connection() as connection:
                async with connection.cursor(aiomysql.cursors.SSCursor) as cur:
                    await cur.execute(self._query)
                    event = {}
                    while True:
                        await self.Pipeline.ready()
                        row = await cur.fetchone()
                        if row is None:
                            break

                        # Transform row to an event object
                        for i, val in enumerate(row):
                            event[cur.description[i][0]] = val

                        # Pass event to the pipeline
                        await self.process(event)
                    await cur.execute("COMMIT;")
        except (
            pymysql.err.InternalError,
            pymysql.err.ProgrammingError,
            pymysql.err.OperationalError,
        ) as e:
            if e.args[0] in self._connection.RetryErrors:
                L.warn(
                    "Recoverable error '{}' occurred in MySQLSource. Retrying.".format(
                        e.args[0]
                    )
                )
            raise e



import logging
import bspump.asab as asab

from .service import MetricsService
from .metrics import (
    Metric,
    Gauge,
    Counter,
    EPSCounter,
    DutyCycle,
    AggregationCounter,
    Histogram,
    CounterWithDynamicTags,
    AggregationCounterWithDynamicTags,
    HistogramWithDynamicTags,
)

#

L = logging.getLogger(__name__)

#


class Module(asab.Module):
    def __init__(self, app):
        super().__init__(app)
        self.service = MetricsService(app, "asab.MetricsService")


__all__ = (
    "MetricsService",
    "Metric",
    "Gauge",
    "Counter",
    "EPSCounter",
    "DutyCycle",
    "AggregationCounter",
    "Histogram",
    "CounterWithDynamicTags",
    "AggregationCounterWithDynamicTags",
    "HistogramWithDynamicTags",
)



from ...abc import Expression
from ..value.valueexpr import VALUE


class CONTAINS(Expression):
    Attributes = {
        "Value": ["str"],
        "Substring": ["str"],
    }

    Category = "String"

    def __init__(self, app, *, arg_what, arg_substring):
        super().__init__(app)
        self.Value = arg_what

        if not isinstance(arg_substring, Expression):
            self.Substring = VALUE(app, value=arg_substring)
        else:
            self.Substring = arg_substring

    def get_outlet_type(self):
        return bool.__name__

    def consult_inlet_type(self, key, child):
        return str.__name__

    def __call__(self, context, event, *args, **kwargs):
        value = self.Value(context, event, *args, **kwargs)
        if value is None:
            return False

        substr = self.Substring(context, event, *args, **kwargs)
        return substr in value



import numpy as np
import zlib


class HyperLogLog(object):
    """
    This is the implementation of HyperLogLog++ algorithm,
    which estimates cardinality of the set with average 2%,
    described in http://algo.inria.fr/flajolet/Publications/FlFuGaMe07.pdf
    and https://storage.googleapis.com/pub-tools-public-publication-data/pdf/40671.pdf
    """

    alphas = {16: 0.673, 32: 0.697, 64: 0.709}

    def __init__(self, m=2048):
        """
        `m` is number of registers. It is bounded with b, b = log2m. Higher `m` is,
        the more precise calculation, however, in the paper above it is proved, that
        `m` = 2048 is optimal and produces error around 2%, other values can deviate
        up to 20%.
        `b` is the number of last bits of the value to take.
        `alpha` is the parameter from papers above.
        """

        self.num_bits = 32
        self.b = int(np.ceil(np.log2(m)))
        self.max = 2**self.num_bits
        self.m = m

        if m >= 128:
            self.alpha = 0.7213 / (1 + 1.079 / m)
        else:
            self.alpha = self.alphas.get(m)

        if self.alpha is None:
            raise RuntimeError(
                "Incorrect m, it should be 16, 32 or 64, or powers of 2 >= 128"
            )

    def add(self, value, array):
        """
        `value` might be string or number.
        `array` is a storage to 'add' the value.
        """

        hashed_value = self.hash_data(value)
        position = self._compute_position(hashed_value)
        rho = self._compute_rho(hashed_value)
        array[position] = np.max([array[position], rho])

    def count(self, array):
        """
        Count unique values in array.
        """

        z = self._compute_z(array)
        e = self._compute_e(z, array)
        return int(e)

    def hash_data(self, value):
        """
        Override it, if you want to use different hash.
        Hash must be 32bit and fast (don't use cryptographic hashes then)
        """

        if not isinstance(value, str):
            value = str(value)

        value = value.encode("utf8")
        return zlib.crc32(value)

    def compute_error(self, ground_truth, hll_count):
        """
        If the ground truth is known, the error can be calculated in %.
        """

        return np.abs(hll_count - ground_truth) / ground_truth * 100

    def _compute_z(self, array):
        twos = [0.5] * len(array)
        z = 1 / float(np.sum(np.power(twos, array)))
        return z

    def _compute_e(self, z, array):
        e = z * self.alpha * self.m**2
        if e < 5 / 2 * self.m:
            v = self._get_zeros(array)
            if v != 0:
                e_star = float(self._linear_count(v))

            else:
                e_star = e

        elif e < 1 / 30 * self.max:
            e_star = e

        else:
            e_star = -self.max * float(np.log(1 - e / self.max))
        return e_star

    def _get_zeros(self, array):
        return len(array[array == 0])

    def _linear_count(self, v):
        return self.m * np.log(self.m / v)

    def _compute_position(self, hashed_value):
        """
        takes b right bits.
        """
        move = hashed_value & (self.max - 1) >> (self.num_bits - self.b)
        return move

    def _compute_rho(self, hashed_value):
        """
        rho = 1 + <leftmost 1 position>
        """
        # SLOW: TODO: find the way to do it better
        x = hashed_value
        pos = self.num_bits
        while x > 1:
            x = x >> 1
            pos -= 1

        return pos



from .application import BSPumpApplication
from .service import BSPumpService
from .pipeline import Pipeline
from .pumpbuilder import PumpBuilder
from .abc.source import Source
from .abc.source import TriggerSource
from .abc.sink import Sink
from .abc.processor import Processor
from .abc.generator import Generator
from .abc.connection import Connection

from .exception import ProcessingError
from .abc.lookup import Lookup
from .abc.lookup import MappingLookup
from .abc.lookup import DictionaryLookup
from .fileloader import load_json_file

from .analyzer.analyzer import Analyzer

from .matrix.matrix import Matrix, PersistentMatrix
from .matrix.namedmatrix import NamedMatrix, PersistentNamedMatrix
from .model.model import Model

from .abc.anomaly import Anomaly

from .__version__ import __version__, __build__

__all__ = (
    "BSPumpApplication",
    "BSPumpService",
    "Pipeline",
    "PumpBuilder",
    "Source",
    "TriggerSource",
    "Sink",
    "Processor",
    "Generator",
    "Connection",
    "Analyzer",
    "ProcessingError",
    "Lookup",
    "MappingLookup",
    "DictionaryLookup",
    "load_json_file",
    "Matrix",
    "PersistentMatrix",
    "NamedMatrix",
    "Model",
    "PersistentNamedMatrix",
    "Anomaly",
    "__version__",
    "__build__",
    "step",
)



from .abcsource import HTTPABCClientSource


class HTTPClientSource(HTTPABCClientSource):
    async def read(self, response):
        response = await response.read()
        await self.process(response)


class HTTPClientTextSource(HTTPABCClientSource):
    ConfigDefaults = {
        "encoding": "",
    }

    def __init__(self, app, pipeline, id=None, config=None, headers={}):
        super().__init__(app, pipeline, id=id, config=config, headers={})

        self.encoding = self.Config["encoding"]
        if self.encoding == "":
            self.encoding = None

    async def read(self, response):
        response = await response.text(encoding=self.encoding)
        await self.process(response)


class HTTPClientLineSource(HTTPClientTextSource):
    async def read(self, response):
        response = await response.text(encoding=self.encoding)

        for line in response.split("\n"):
            await self.process(line)



import datetime
import pytz

from bspump.asab import Config

from ...abc import Expression, evaluate


class DATETIME_GET(Expression):
    Attributes = {
        "Value": ["*"],  # TODO: This ...
        "What": ["*"],  # TODO: This ...
        "Timezone": ["*"],  # TODO: This ...
    }

    Category = "Date/Time"

    def __init__(self, app, *, arg_with, arg_what, arg_timezone=None):
        super().__init__(app)

        self.Value = arg_with
        self.What = arg_what

        if arg_what in (
            "year",
            "month",
            "day",
            "hour",
            "minute",
            "second",
            "microsecond",
        ):
            self.Method = 1
            self.What = arg_what

        elif arg_what in ("timestamp", "weekday", "isoweekday"):
            self.Method = 2
            self.What = arg_what

        else:
            raise ValueError("Invalid 'what' provided: '{}'".format(arg_what))

        if arg_timezone is None:
            timezone_from_config = Config["declarations"]["timezone"]

            if len(timezone_from_config) == 0:
                self.Timezone = None

            else:
                self.Timezone = pytz.timezone(timezone_from_config)

        else:
            self.Timezone = pytz.timezone(arg_timezone)

    def __call__(self, context, event, *args, **kwargs):
        value = evaluate(self.Value, context, event, *args, **kwargs)
        if isinstance(value, int) or isinstance(value, float):
            value = datetime.datetime.utcfromtimestamp(value)

        # Apply the timezone
        if self.Timezone is not None:
            value = self.Timezone.localize(value)

        if self.Method == 1:
            try:
                return getattr(value, self.What)
            except AttributeError:
                return None

        elif self.Method == 2:
            func = getattr(value, self.What)
            return func()



from .fileblocksink import FileBlockSink
from .fileblocksource import FileBlockSource
from .filecsvsink import FileCSVSink
from .filecsvsource import FileCSVSource
from .filejsonsource import FileJSONSource
from .filelinesource import FileLineSource
from .filelinesource import FileMultiLineSource
from .lookupprovider import FileBatchLookupProvider

__all__ = (
    "FileBlockSink",
    "FileBlockSource",
    "FileCSVSink",
    "FileCSVSource",
    "FileJSONSource",
    "FileLineSource",
    "FileMultiLineSource",
    "FileBatchLookupProvider",
)



import collections.abc

from ..abc.processor import Processor
from ..abc.generator import Generator


class MappingKeysProcessor(Processor):
    """
    Description: Mapping Keys Processor

    |

    """

    def process(self, context, event: collections.abc.Mapping) -> list:
        """
        Description: process is a method of a Mapping Keys Processor

        :return: event.keys()

        |

        """
        return [*event.keys()]


class MappingValuesProcessor(Processor):
    """
    Description:

    |

    """

    def process(self, context, event: collections.abc.Mapping) -> list:
        """
        Description:

        :return: event.values()

        |

        """
        return [*event.values()]


class MappingItemsProcessor(Processor):
    """
    Description:

    |

    """

    def process(self, context, event: collections.abc.Mapping) -> list:
        """
        Description:

        :return: event.items()

        |

        """
        return [*event.items()]


class MappingKeysGenerator(Generator):
    """
    Description:

    |

    """

    async def generate(self, context, event, depth):
        """
        Description:

        """
        for item in event.keys():
            self.Pipeline.inject(context, item, depth)


class MappingValuesGenerator(Generator):
    """
    Description:

    |

    """

    async def generate(self, context, event, depth):
        """
        Description:

        |

        """
        for item in event.values():
            self.Pipeline.inject(context, item, depth)


class MappingItemsGenerator(Generator):
    """
    Description:

    |

    """

    async def generate(self, context, event, depth):
        """
        Description:

        |

        """
        for item in event.items():
            self.Pipeline.inject(context, item, depth)



import time
import logging
import collections

import asyncio

from bspump.asab import Configurable

from .generalanomaly import GeneralAnomaly

###

L = logging.getLogger(__name__)

###


class AnomalyStorage(Configurable, collections.OrderedDict):
    """
    AnomalyStorage serves to store anomaly objects (see AnomalyManager for details),
    separated to "open" (anomalies that are not closed by status attribute in a symptom) and "closed".

    The closed anomalies are periodically flushed from the storage to an external system.
    """

    ConfigDefaults = {
        "closed_anomaly_longevity": 5,  # Anomalies come pretty often
        "index": "bs_anomaly*",
    }

    def __init__(
        self,
        app,
        es_connection,
        anomaly_classes=list(),
        anomaly_storage_pipeline_source="AnomalyStoragePipeline.*InternalSource",
        pipeline=None,
        id="AnomalyStorage",
        config=None,
    ):
        super().__init__(config_section_name=id, config=config)

        self["open"] = {}
        self["closed"] = {}

        self.App = app
        self.Id = id
        self.Pipeline = pipeline
        self.AnomalyStoragePipelineSource = anomaly_storage_pipeline_source
        self.AnomalyClasses = anomaly_classes
        self.ClosedAnomalyLongevity = int(self.Config["closed_anomaly_longevity"])
        self.Index = str(self.Config["index"])
        self.Connection = es_connection

        # Subscribe to periodically flush old closed anomalies
        self.App.PubSub.subscribe("Application.tick/300!", self.flush)

        metrics_service = self.App.get_service("asab.MetricsService")
        self.AnomalyStorageCounter = metrics_service.create_counter(
            "anomaly.storage",
            init_values={
                "anomalies.closed.flushed": 0,
                "anomalies.open.flushed": 0,
            },
        )

        # Load previously saved anomalies from the external system
        self.LoadTasks = [asyncio.ensure_future(self.load())]
        self.App.PubSub.subscribe("Application.exit!", self._on_exit)

    def set_pipeline(self, pipeline):
        self.Pipeline = pipeline

    async def load(self):
        query = {
            "size": 10000,
            "query": {
                "bool": {
                    "must": [
                        {
                            "exists": {"field": "@timestamp"},
                        },
                        {
                            "exists": {"field": "type"},
                        },
                        {
                            "exists": {"field": "status"},
                        },
                        {
                            "match": {
                                "status": {
                                    "query": "open",
                                    "minimum_should_match": 1,
                                    "zero_terms_query": "all",
                                }
                            }
                        },
                        {
                            "range": {
                                "@timestamp": {"gte": "now-14d", "lte": "now"},
                            },
                        },
                    ]
                }
            },
            "sort": [{"@timestamp": {"order": "asc"}}],
        }

        url = self.Connection.get_url() + "{}/_search".format(self.Index)
        async with self.Connection.get_session() as session:
            async with session.post(
                url, json=query, headers={"Content-Type": "application/json"}
            ) as response:
                if response.status != 200:
                    data = await response.text()
                    L.error(
                        "Failed to fetch data from ElasticSearch: {} from {}\n{}".format(
                            response.status, url, data
                        )
                    )
                    return
                msg = await response.json()

        hits = msg["hits"]["hits"]
        if len(hits) == 0:
            L.warning("No open anomalies present in ElasticSearch ...")
            return

        # Save open/future anomalies to the in-memory dictionary
        for hit in hits:
            key = hit["_id"]
            # Prevent backups
            if not key.startswith("b_"):
                event = hit["_source"]
                # Modify timestamp
                event["@timestamp"] = int(event["@timestamp"] / 1000)
                # Select proper anomaly class
                selected_anomaly_class = GeneralAnomaly
                for anomaly_class in self.AnomalyClasses:
                    if anomaly_class.TYPE == event["type"]:
                        selected_anomaly_class = anomaly_class
                        break
                # Add all data
                anomaly = selected_anomaly_class()
                for a_key, a_value in event.items():
                    anomaly[a_key] = a_value
                # Save the anomaly
                self["open"][key] = anomaly

        L.info("Open anomalies loaded ...")

    async def _on_exit(self, message_type):
        await asyncio.wait(self.LoadTasks)

    async def flush(self, message_type):
        L.info("Start flushing of closed anomalies ...")

        current_time = int(time.time())

        # Throttle the parental pipeline
        if self.Pipeline is not None:
            self.Pipeline.throttle(self.Id, True)

        svc = self.App.get_service("bspump.PumpService")
        anomaly_storage_pipeline_source = svc.locate(self.AnomalyStoragePipelineSource)

        if anomaly_storage_pipeline_source is None:
            L.warning("The anomaly storage pipeline is not yet ready, skipping ...")
            return

        keys_to_be_deleted = []

        # Update anomalies & close them, if it is possible
        for key, anomaly in self["open"].items():
            await anomaly.on_tick(current_time)
            if anomaly["status"] == "closed":
                anomaly.close(current_time)
                self["closed"][key] = anomaly
                keys_to_be_deleted.append(key)

        # Remove moved anomalies from the storage
        for key in keys_to_be_deleted:
            del self["open"][key]

        keys_to_be_deleted = []

        # FLUSH old closed anomalies to external database
        await asyncio.sleep(0.01)
        # Synchronous to make atomic cycle
        for key, anomaly in self["closed"].items():
            # Pass anomaly to the storage pipeline
            context = {
                "es_id": key,
            }
            await anomaly_storage_pipeline_source.put_async(context, anomaly)
            # Delete the closed anomaly after some time
            if current_time > anomaly["ts_end"] + self.ClosedAnomalyLongevity:
                self.AnomalyStorageCounter.add("anomalies.closed.flushed", 1)
                keys_to_be_deleted.append(key)

        # Remove old closed anomalies from the storage
        for key in keys_to_be_deleted:
            del self["closed"][key]

        # FLUSH open anomalies for persistence to external system
        await asyncio.sleep(0.01)
        # Synchronous to make atomic cycle
        for key, anomaly in self["open"].items():
            context = {
                "es_id": key,
            }
            await anomaly_storage_pipeline_source.put_async(context, anomaly)
            self.AnomalyStorageCounter.add("anomalies.open.flushed", 1)

        # Throttle the parental pipeline
        if self.Pipeline is not None:
            self.Pipeline.throttle(self.Id, False)

        L.info("End flushing of closed anomalies ...")



#!/usr/bin/env python3
import logging
import time

import bspump
import bspump.common
import bspump.elasticsearch
import bspump.file
import bspump.trigger

###

L = logging.getLogger(__name__)

###


class LoadSource(bspump.TriggerSource):
    def __init__(self, app, pipeline, choice=None, id=None, config=None):
        super().__init__(app, pipeline, id=id, config=config)
        self.Number = 200000

    async def cycle(self):
        print("START ----")
        stime = time.time()
        for i in range(0, self.Number):
            event = {"name": "Chuck Norris"}
            await self.process(event)
        etime = time.time()
        print("EPS: {:.0f}".format(self.Number / (etime - stime)))


class SamplePipeline(bspump.Pipeline):
    def __init__(self, app, pipeline_id):
        super().__init__(app, pipeline_id)
        self.build(
            LoadSource(app, self).on(
                bspump.trigger.OpportunisticTrigger(app, chilldown_period=10)
            ),
            bspump.common.NullSink(app, self),
        )


if __name__ == "__main__":
    app = bspump.BSPumpApplication()

    svc = app.get_service("bspump.PumpService")

    # Construct and register Pipeline
    pl = SamplePipeline(app, "SamplePipeline")
    svc.add_pipeline(pl)

    app.run()



from .index import Index, BitMapIndex, TreeRangeIndex, SliceIndex
from .ipgeolookup import IPGeoLookup
from .matrixlookup import MatrixLookup

__all__ = (
    "IPGeoLookup",
    "MatrixLookup",
    "Index",
    "BitMapIndex",
    "TreeRangeIndex",
    "SliceIndex",
)



from ..abc.source import TriggerSource


class CronSource(TriggerSource):
    """
    A simplified source for cron-triggered pipelines.

    Usage:
        auto_pipeline(
            source=lambda app, pipeline: CronSource(
                app, pipeline,
                config={"when": "*/10 * * * *"}
            ),
            sink=lambda app, pipeline: PPrintSink(app, pipeline),
        )

    Config:
        when (str): Cron expression (e.g., "0 9 * * *" for daily at 9 AM)
        init_time (datetime, optional): Initial time for cron calculation (naive)
    """

    def __init__(self, app, pipeline, id=None, config=None):
        super().__init__(app, pipeline, id=id, config=config)

        cron_string = self.Config.get("when")
        if not cron_string:
            raise ValueError(
                "CronSource requires 'when' config parameter with cron expression"
            )

        init_time = self.Config.get("init_time")
        if init_time is None:
            from datetime import datetime

            init_time = datetime.now()  # naive for backward compatibility

        from ..trigger import CronTrigger

        cron_trigger = CronTrigger(app, cron_string, init_time, id=f"{self.Id}_cron")
        self.on(cron_trigger)

    async def cycle(self, *args, **kwargs):
        await self.Pipeline.ready()
        from datetime import datetime

        now = datetime.now()  # naive
        event = {
            "cron_triggered": now.isoformat(),
            "trigger_id": self.Id,
            "timestamp": now.timestamp(),
        }
        await self.Pipeline.process(event)



from bspump.jupyter import *
from bspump.test import TestSink, TestSource
import json
import asyncio


foovar=3


event = {}
test_events = [({"foo":"aaa"},{"expect" : [{"foo": "aaa", "foovar": 3, "barvar": "hello", "bazvar": "aaa"}]}), 
               ({"foo":"aab"},{"expect" : [{"foo": "aab", "foovar": 3, "barvar": "hello", "bazvar": "aab"}]})]


auto_pipeline(
    source=lambda app, pipeline: TestSource(app, pipeline, "TestSource"),
    sink=lambda app, pipeline: TestSink(app, pipeline, "TestSink")
)


@async_step
async def processor_internal(inject, event):
    barvar="hello"
    await asyncio.sleep(0.2)
    my_multiline_string = """a line
non-indented line
    once indented line
        twice indented line
"""
    bazvar=event["foo"]


    print("Next step")
    event["foovar"] = foovar
    event["barvar"] = barvar
    event["bazvar"] = bazvar
    event


    await inject(event)



from ..abc.processor import Processor


class StringToBytesParser(Processor):
    """
    Description:

    |
    ** Default Config **

    encoding : utf-8

    """

    ConfigDefaults = {
        "encoding": "utf-8",
    }

    def __init__(self, app, pipeline, id=None, config=None):
        """
        Description:

        |

        """
        super().__init__(app, pipeline, id, config)
        self.Encoding = self.Config["encoding"]

    def process(self, context, event):
        """
        Description:

        :return: event.decode(self.Encoding)

        |

        """
        assert isinstance(event, str)
        return event.encode(self.Encoding)


class BytesToStringParser(Processor):
    """
    Description:

    |

    """

    ConfigDefaults = {
        "encoding": "utf-8",
    }

    def __init__(self, app, pipeline, id=None, config=None):
        """
        Description:

        |

        """
        super().__init__(app, pipeline, id, config)
        self.Encoding = self.Config["encoding"]

    def process(self, context, event):
        """
        Description:

        :return: event.decode(self.Encoding)

        |

        """
        assert isinstance(event, bytes)
        return event.decode(self.Encoding)



import logging
from .routing import InternalSource, RouterProcessor


L = logging.getLogger(__name__)


class TeeSource(InternalSource):
    """
            Description:



    class SamplePipeline(bspump.Pipeline):

            def __init__(self, app, pipeline_id):
                    super().__init__(app, pipeline_id)

                    self.build(
                            bspump.socket.TCPStreamSource(app, self, config={'port': 7000}),
                            bspump.common.TeeProcessor(app, self).bind("SampleTeePipeline.*TeeSource"),
                            bspump.common.PPrintSink(app, self)
                    )


    class SampleTeePipeline(bspump.Pipeline):

            def __init__(self, app, pipeline_id):
                    super().__init__(app, pipeline_id)

                    self.build(
                            bspump.common.TeeSource(app, self),
                            bspump.common.PPrintSink(app, self)
                    )

            |

    """

    def __init__(self, app, pipeline, id=None, config=None):
        """
        Description:

        |

        """
        super().__init__(app, pipeline, id=id, config=config)

        self.Targets = []
        self._svc = app.get_service("bspump.PumpService")

    def bind(self, target):
        """
        Description:

        :return:

        |

        """
        self.Targets.append(target)
        return self

    async def main(self):
        """
        Description:

        :return:

        |

        """

        unbind_processor = []
        for target in self.Targets:
            processor = self._svc.locate(target)
            if processor is None:
                L.warning(
                    "TeeSource '{}' cannot find processor '{}'".format(self.Id, target)
                )
                return

            if not isinstance(processor, TeeProcessor):
                L.warning(
                    "TeeSource '{}' requires TeeProcessor as target, not '{}'".format(
                        self.Id, target
                    )
                )
                return

            processor.bind(self.locate_address())
            unbind_processor.append(processor)

        try:
            await super().main()
        finally:
            for processor in unbind_processor:
                processor.unbind(self.locate_address())


#


class TeeProcessor(RouterProcessor):
    """
    Description: See TeeSource for details.

    |

    """

    ConfigDefaults = {}

    def __init__(self, app, pipeline, id=None, config=None):
        """
        Description:

        |

        """
        super().__init__(app, pipeline, id=id, config=config)
        self.Targets = []

    def bind(self, target: str):
        """
        Description:	Target is a bspump.PumpService.locate() string

        :return: ?

        |

        """

        self.Targets.append(target)
        return self

    def unbind(self, target: str):
        """
        Description:

        :return: ?

        |

        """
        self.Targets.remove(target)
        self.unlocate(target)
        return self

    def process(self, context, event):
        """
        Description:

        :return: event

        |

        """
        for source in self.Targets:
            self.route(context, event, source)
        return event



import logging
import json
from .fileabcsource import FileABCSource


L = logging.getLogger(__file__)


class FileJSONSource(FileABCSource):
    """
    Description: This file source is optimized to load even large JSONs from a file and parse that.
    The loading & parsing is off-loaded to the worker thread so that it doesn't block the IO loop.

    """

    def __init__(self, app, pipeline, id=None, config=None):
        """
        Description:

        **Parameters**

        app :

        pipeline :

        id : ID, default= None
                ID

        config : JSON, default = None
                configuration file with additional information

        """
        super().__init__(app, pipeline, id=id, config=config)
        self.ProactorService = app.get_service("asab.ProactorService")

    async def read(self, filename, f):
        """
        Description:

        **Parameters**

        filename :

        f :

        """
        await self.Pipeline.ready()

        worker = self.ProactorService.execute(json.load, f)
        await worker
        event = worker.result()
        await self.process(event, {"filename": filename})



from .pipeline import UnitTestPipeline
from .sink import UnitTestSink
from .source import UnitTestSource
from .unit_test_case import ProcessorTestCase
from .unit_test_case import TestCase

__all__ = (
    "UnitTestPipeline",
    "UnitTestSink",
    "UnitTestSource",
    "ProcessorTestCase",
    "TestCase",
)



import logging
import asyncio
from ..abc.sink import Sink


L = logging.getLogger(__name__)

# TODO: Research option to refactor with aiogoogle
# https://aiogoogle.readthedocs.io/en/latest/


class GoogleDriveABCSink(Sink):
    """
    GoogleDriveABCSink is abstract class ment to be used for uploading files to GoogleDrive.
    """

    def __init__(self, app, pipeline, connection, id=None, config=None):
        super().__init__(app, pipeline, id=id, config=config)

        self.Drive_service = None
        self._conn_future = None
        self.Connection = pipeline.locate_connection(app, connection)
        self._output_queue = asyncio.Queue(
            maxsize=self.Connection._output_queue_max_size + 1
        )

        # Subscription
        app.PubSub.subscribe("Application.stop!", self._on_exit)
        app.PubSub.subscribe("Application.tick!", self._on_health_check)
        self._on_health_check("connection.open!")

    def _on_health_check(self, message_type):
        if self._conn_future is not None:
            # Connection future exists

            if not self._conn_future.done():
                # Connection future didn't result yet
                # No sanitization needed
                return

            try:
                self._conn_future.result()
            except Exception as e:
                # Connection future threw an error
                L.exception(f"Unexpected connection future error {e}")

            # Connection future already resulted (with or without exception)
            self._conn_future = None

        assert self._conn_future is None

        self._conn_future = asyncio.ensure_future(
            self._connection(),
        )

    async def _connection(self):
        # producer = await self.Connection.create_producer(**self._producer_params)
        if self.Drive_service is None:
            self.Drive_service = self.Connection._get_service()
        await self._loader()

    def process(self, context, event):
        self._output_queue.put_nowait((event, context))

        if self._output_queue.qsize() == self.Connection._output_queue_max_size:
            self.Pipeline.throttle(self, True)

    async def _on_exit(self, event_name, counter=None):
        self._output_queue.put_nowait((None, None))

    async def _loader(self):
        raise NotImplementedError()



import dataclasses


@dataclasses.dataclass
class LibraryItem:
    """
    The data class that contains the info about a specific item in the library.

    Attributes:
            name (str): The absolute path of the Item. It can be directly fed into `LibraryService.read(...)`.
            type (str): Can be either `dir` if the Item is a directory or `item` if Item is of any other type.
            layer (int): The number of highest layer in which this Item is found. The higher the number, the lower the layer is.
            providers (list): List of `LibraryProvider` objects containing this Item.
            disabled (bool): `True` if the Item is disabled, `False` otherwise. If the Item is disabled, `LibraryService.read(...)` will return `None`.
            override (int): If `True`, this item is marked as an override for the providers with the same Item name.
    """

    name: str
    type: str
    layer: int
    providers: list
    disabled: bool = False
    override: int = 0  # Default value for override is False



import typing
import aiohttp.web

from ..abc import Module
from .container import WebContainer
from .websocket import WebSocketFactory
from .staticdir import StaticDirProvider


class Module(Module):
    """
    Module for running and easy manipulation of the web server.
    """

    def __init__(self, app):
        super().__init__(app)

        from .service import WebService

        self.service = WebService(app, "asab.WebService")


def create_web_server(
    app, section: str = "web", config: typing.Optional[dict] = None
) -> aiohttp.web.UrlDispatcher:
    """
    Build the web server with the specified configuration.

    It is an user convenience function that simplifies typical way of how the web server is created.

    Args:
            app (asab.Application): A reference to the ASAB Application.
            section (str): Configuration section name with which the WebContainer will be created.
            config (dict | None): Additional server configuration.

    Returns:
            [WebContainer Application Router object](https://docs.aiohttp.org/en/stable/web_reference.html?highlight=router#router).

    Examples:

    ```python
    class MyApplication(asab.Application):
            async def initialize(self):
                    web = asab.web.create_web_server(self)
                    web.add_get('/hello', self.hello)

            async def hello(self, request):
                    return asab.web.rest.json_response(request, data="Hello, world!\n")

    ```
    """
    app.add_module(Module)
    websvc = app.get_service("asab.WebService")
    container = WebContainer(websvc, section, config=config)
    return container.WebApp.router


__all__ = (
    "WebContainer",
    "WebSocketFactory",
    "StaticDirProvider",
    "Module",
)



import asyncio
import logging

import aioodbc

from asab import PubSub
from ..abc.connection import Connection

#

L = logging.getLogger(__name__)

#


class ODBCConnection(Connection):
    # Caution: Providing incorrect connection configuration terminates the program with 'Abort trap 6'
    ConfigDefaults = {
        "dsn": "",
        "host": "",
        "port": -1,
        "user": "",
        "password": "",
        "connect_timeout": 1,
        "reconnect_delay": 5.0,
        "output_queue_max_size": 10,
        "max_bulk_size": 2,
    }

    def __init__(self, app, id=None, config=None):
        super().__init__(app, id=id, config=config)
        self.ConnectionEvent = asyncio.Event()
        self.ConnectionEvent.clear()

        self.PubSub = PubSub(app)
        self.Loop = app.Loop

        self._dsn = self.Config["dsn"]
        self._host = self.Config["host"]
        self._port = int(self.Config["port"])
        self._user = self.Config["user"]
        self._password = self.Config["password"]

        self._connect_timeout = self.Config["connect_timeout"]
        self._reconnect_delay = self.Config["reconnect_delay"]
        self._output_queue_max_size = self.Config["output_queue_max_size"]
        self._max_bulk_size = int(self.Config["max_bulk_size"])

        self._conn_future = None
        self._connection_request = False
        self._pause = False

        # Subscription
        self._on_health_check("connection.open!")
        app.PubSub.subscribe("Application.stop!", self._on_application_stop)
        app.PubSub.subscribe("Application.tick!", self._on_health_check)
        app.PubSub.subscribe("ODBCConnection.pause!", self._on_pause)
        app.PubSub.subscribe("ODBCConnection.unpause!", self._on_unpause)

        self._output_queue = asyncio.Queue()
        self._bulks = {}  # We have a "bulk" per query

    def _on_pause(self):
        self._pause = True

    def _on_unpause(self):
        self._pause = False

    def _flush(self):
        for query in self._bulks.keys():
            # Break if throttling was requested during the flush,
            # so that put_nowait doesn't raise
            if self._pause:
                break

            self._flush_bulk(query)

    def _flush_bulk(self, query):
        # Enqueue and thorttle if needed
        self._output_queue.put_nowait((query, self._bulks[query]))
        if self._output_queue.qsize() == self._output_queue_max_size:
            self.PubSub.publish("ODBCConnection.pause!", self)

        # Reset bulk
        self._bulks[query] = []

    def _on_application_stop(self, message_type, counter):
        self._flush()
        self._output_queue.put_nowait((None, None))

    def _on_health_check(self, message_type):
        if self._conn_future is not None:
            # Connection future exists

            if not self._conn_future.done():
                # Connection future didn't result yet
                # No sanitization needed
                return

            try:
                self._conn_future.result()
            except Exception:
                # Connection future threw an error
                L.exception("Unexpected connection future error")

            # Connection future already resulted (with or without exception)
            self._conn_future = None

        assert self._conn_future is None

        self._conn_future = asyncio.ensure_future(
            self._connection(),
        )

    async def _connection(self):
        kwargs = {}

        if len(self._host) != 0:
            kwargs["host"] = self._host

        if self._port != -1:
            kwargs["port"] = self._port

        if len(self._user) != 0:
            kwargs["user"] = self._user

        if len(self._password) != 0:
            kwargs["password"] = self._password

        if len(self._dsn) != 0:
            kwargs["dsn"] = self._dsn

        try:
            async with aioodbc.create_pool(
                connect_timeout=self._connect_timeout, loop=self.Loop, **kwargs
            ) as pool:
                self._conn_pool = pool
                self.ConnectionEvent.set()
                await self._loader()

        except BaseException:
            L.exception("Unexpected ODBC connection error")
            raise

    def acquire(self):
        assert self._conn_pool is not None
        return self._conn_pool.acquire()

    def consume(self, query, data):
        # Create a bulk for this query if doesn't yet exist
        if query not in self._bulks:
            self._bulks[query] = []

        # Add data to the query's bulk
        self._bulks[query].append(data)

        # Flush on _max_bulk_size
        if len(self._bulks[query]) >= self._max_bulk_size:
            self._flush_bulk(query)

    async def _loader(self):
        while True:
            query, data = await self._output_queue.get()
            if query is None:
                break

            if self._output_queue.qsize() == self._output_queue_max_size - 1:
                self.PubSub.publish(
                    "ODBCConnection.unpause!", self, asynchronously=True
                )
            async with self.acquire() as conn:
                async with conn.cursor() as cur:
                    await cur.executemany(query, data)
                    await conn.commit()



from ...abc import SequenceExpression, evaluate


class FIRST(SequenceExpression):
    Category = "Statements"

    def __call__(self, context, event, *args, **kwargs):
        for item in self.Items:
            res = evaluate(item, context, event, *args, **kwargs)
            if res is not None:
                return res



import abc

import bspump.asab as asab


class LookupProviderABC(abc.ABC, asab.Configurable):
    """
    Description:

    |

    """

    def __init__(self, lookup, url, id=None, config=None):
        """
        Description:

        """
        self.Id = "lookupprovider:{}".format(
            id if id is not None else self.__class__.__name__
        )
        super().__init__(config_section_name=self.Id, config=config)
        self.Lookup = lookup
        self.App = lookup.App
        self.URL = url
        self.ETag = None

    async def load(self):
        """
        Description:

        |

        """
        raise NotImplementedError()


class LookupBatchProviderABC(LookupProviderABC, abc.ABC):
    """
    Description:

    |

    """

    pass



from ..abc.sink import Sink


class SlackAbcSink(Sink):
    def __init__(self, app, pipeline, connection, id=None, config=None):
        super().__init__(app, pipeline, id=id, config=config)

        self.Connection = pipeline.locate_connection(app, connection)

        app.PubSub.subscribe("SlackConnection.pause!", self._connection_throttle)
        app.PubSub.subscribe("SlackConnection.unpause!", self._connection_throttle)

    def _connection_throttle(self, event_name, connection):
        if connection != self.Connection:
            return

        if event_name == "SlackConnection.pause!":
            self.Pipeline.throttle(self, True)
        elif event_name == "SlackConnection.unpause!":
            self.Pipeline.throttle(self, False)
        else:
            raise RuntimeError("Unexpected event name '{}'".format(event_name))


class SlackTextSink(SlackAbcSink):
    def process(self, context, event):
        """
        Param 'event' must be a string
        """
        msg = {"text": str(event)}
        self.Connection.consume(msg)


class SlackMessageSink(SlackAbcSink):
    def process(self, context, event):
        """
        Param 'event' must be a Slack message (https://api.slack.com/docs/messages)
        It must contains at least keys 'text' or 'attachments'
        Example:
                {
                        "text": "This is text example!",
                        "attachments": [
                                {
                                        "title": "This is attachments title",
                                        "fields": [
                                                {
                                                        "title": "Volume",
                                                        "value": "1",
                                                        "short": true
                                                }
                                        ]
                                }
                        ]
                }
        """

        self.Connection.consume(event)



import time

import bspump
import bspump.analyzer
import bspump.unittest


class TestTimeWindowMatrix(bspump.unittest.TestCase):
    def test_matrix(self):
        matrix = bspump.matrix.TimeWindowMatrix(app=self.App, clock_driven=False)
        self.assertEqual(matrix.Array.shape, (1, 15))

    def test_matrix_add_column(self):
        matrix = bspump.matrix.TimeWindowMatrix(
            app=self.App, columns=3, clock_driven=False
        )
        start = matrix.TimeConfig.get_start()
        end = matrix.TimeConfig.get_end()
        row_index = matrix.add_row("abc")
        warming_up = matrix.WarmingUpCount.WUC[row_index]
        first_col = 4
        second_col = 5
        third_col = 6
        matrix.Array[row_index, 0] = first_col
        matrix.Array[row_index, 1] = second_col
        matrix.Array[row_index, 2] = third_col
        num_columns = matrix.Array.shape[1]
        matrix.add_column()

        self.assertEqual(matrix.TimeConfig.get_start(), start + matrix.Resolution)
        self.assertEqual(matrix.TimeConfig.get_end(), end + matrix.Resolution)
        self.assertEqual(matrix.WarmingUpCount.WUC[0], warming_up - 1)

        self.assertEqual(matrix.Array[row_index, 0], second_col)
        self.assertEqual(matrix.Array[row_index, 1], third_col)
        self.assertEqual(matrix.Array[row_index, 2], 0)
        self.assertEqual(matrix.Array.shape[1], num_columns)

    def test_matrix_add_row(self):
        matrix = bspump.matrix.TimeWindowMatrix(
            app=self.App, columns=3, clock_driven=False
        )
        row_index = matrix.add_row("abc")
        self.assertEqual(matrix.Array.shape[0], len(matrix.WarmingUpCount))

    # def test_matrix_close_row(self):
    # 	matrix = bspump.matrix.TimeWindowMatrix(app=self.App, columns=3, clock_driven=False)
    # 	row_index = matrix.add_row("abc")
    # 	self.assertFalse(row_index in matrix.ClosedRows.CR)
    # 	matrix.close_row(row_index)
    # 	self.assertTrue(row_index in matrix.ClosedRows.CR)
    # 	matrix.flush()
    # 	self.assertEqual(0, matrix.Array.shape[0])

    def test_matrix_get_column(self):
        columns = 10
        matrix = bspump.matrix.TimeWindowMatrix(
            app=self.App, resolution=1, columns=columns, clock_driven=False
        )
        cur_time = int(time.time())
        for i in range(columns):
            time_ = cur_time - i
            column_index = matrix.get_column(time_ + 0.5)
            self.assertEqual(column_index, columns - i - 1)

    def test_matrix_advance(self):
        columns = 10
        cur_time = int(time.time())
        matrix = bspump.matrix.TimeWindowMatrix(
            app=self.App,
            start_time=cur_time,
            resolution=1,
            columns=columns,
            clock_driven=False,
        )
        row_index = matrix.add_row("abc")

        target_ts = matrix.TimeConfig.get_start()
        added = matrix.advance(target_ts)
        self.assertGreater(added, 0)

        target_ts = matrix.TimeConfig.get_start() + matrix.Resolution
        added = matrix.advance(target_ts)
        self.assertGreater(added, 0)

        target_ts = matrix.TimeConfig.get_start() - matrix.Resolution
        added = matrix.advance(target_ts)
        self.assertEqual(added, 0)

        target_ts = matrix.TimeConfig.get_start() - 0.5 * matrix.Resolution
        added = matrix.advance(target_ts)
        self.assertEqual(added, 0)

        target_ts = matrix.TimeConfig.get_start() + 0.5 * matrix.Resolution
        added = matrix.advance(target_ts)
        self.assertGreater(added, 0)



import collections


class LRUCacheDict(collections.OrderedDict):
    """
    LRUCacheDict implements the "Least recently used" cache strategy.
    LRUCacheDict removes last used elements, if the time they were lastly used exceeds the specified `max_duration` or the cache dictionary exceeds `max_size`.
    For more information, please see: https://en.wikipedia.org/wiki/Cache_replacement_policies#Least_recently_used_(LRU)

    The following example illustrates how to use LRUCacheDict with MySQLLookup:

            self.MySQLLookup =  MySQLLookup(self,
                    connection=mysql_connection,
                    id="MySQLLookup",
                    cache=bspump.cache.LRUCacheDict(app, max_size=1000, max_duration=1000)
            )

    """

    def __init__(self, app, max_size=1000, max_duration=None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.App = app

        self.MaxSize = max_size
        self.MaxDuration = max_duration
        self.refresh()

    def refresh(self):
        # Remove items that overflow the dictionary
        if self.MaxSize:
            for i in range(max(0, len(self) - self.MaxSize)):
                self.popitem(last=False)

        # Remove items by max time
        if self.MaxDuration:
            limit = self.App.time() - self.MaxDuration
            while self:
                value, lru = next(iter(super().values()))
                if lru > limit:
                    break
                self.popitem(last=False)

    def __getitem__(self, key):
        value = super().__getitem__(key)[0]
        super().__setitem__(key, (value, self.App.time()))
        self.move_to_end(key)
        return value

    def __setitem__(self, key, value):
        super().__setitem__(key, (value, self.App.time()))
        self.refresh()

    def items(self):
        return ((key, value) for key, (value, _) in super().items())

    def values(self):
        return (value for value, _ in super().values())



import asyncio
import logging
import pymongo

from ..abc.source import Source

#

L = logging.getLogger(__name__)

#


class MongoDBChangeStreamSource(Source):
    """
    `MongoDBChangeStreamSource` listens to the specified `Database`
    and (optionally) `Collection` (if not configured, the events are aggregated
    from all collections). The output are `update`, `insert`, `delete`,
    `invalidate`, `dropDatabase`, `drop`, `rename`, `replace` events.
    Examples of events you can find here: https://docs.mongodb.com/manual/reference/change-events/
    WARNING! Make sure, that version of MongoDB is >= 4.0.0 and
    replica set is enabled.
    """

    ConfigDefaults = {
        "database": "",
        "collection": "",
        "full_document": "",
    }

    def __init__(self, app, pipeline, connection, id=None, config=None):
        super().__init__(app, pipeline, id=id, config=config)
        self.Connection = pipeline.locate_connection(app, connection)
        self.Database = self.Config["database"]
        self.Collection = self.Config["collection"]
        self.FullDocument = self.Config["full_document"]
        if self.FullDocument == "":
            self.FullDocument = None
        if self.Collection == "":
            self.Collection = None

    async def main(self):
        running = True
        await self.Pipeline.ready()
        db = self.Connection.Client[self.Database]
        if self.Collection is None:
            if self.FullDocument is None:
                stream = db.watch()
            else:
                stream = db.watch(full_document=self.FullDocument)
        else:
            if self.FullDocument is None:
                stream = db[self.Collection].watch()
            else:
                stream = db[self.Collection].watch(full_document=self.FullDocument)

        while True:
            if not running:
                await stream.close()
                break
            try:
                event = await stream.try_next()
                if event is not None:
                    await self.process(event, context={})

            except asyncio.CancelledError as e:
                L.warning(
                    f"Mongo change stream task cancled or timed out: {e}. Restarting connection."
                )
                return await self.main()
            except pymongo.errors.OperationFailure as e:
                L.warning(
                    f"Recoverable error encountered while reading changestream: {e}. Restarting connection."
                )
                return await self.main()



import asyncio
import logging

from .fileabcsource import FileABCSource


L = logging.getLogger(__file__)


class FileBlockSource(FileABCSource):
    """
    Description:

    """

    def __init__(self, app, pipeline, id=None, config=None):
        """
        Description:

        **Parameters**

        app : Application
                Name of the Application.

        pipeline : Pipeline
                Name of the Pipeline.

        id : ID, default = None
                ID

        config : JSON, default = None
                Configuration file with additional information.
        """
        super().__init__(app, pipeline, id=id, config=config)
        self.ProactorService = app.get_service("asab.ProactorService")

    async def read(self, filename, f):
        """
        Loads a file.

        **Parameters**

        filename : file
                Name of the file.

        f :

        """
        await self.Pipeline.ready()
        # Load the file in a worker thread (to prevent blockage of the main loop)
        worker = self.ProactorService.execute(f.read)

        try:
            await worker
        except asyncio.CancelledError:
            return

        event = worker.result()
        await self.process(event, {"filename": filename})



import ssl
import socket
import asyncio
import logging

from ..abc.source import Source

from .stream import Stream, TLSStream
from .protocol import LineSourceProtocol

#

L = logging.getLogger(__name__)

#


class StreamServerSource(Source):
    """
    Description:

    """

    ConfigDefaults = {
        "address": "127.0.0.1 8888",  # IPv4, IPv6 or unix socket path
        "backlog": "",
        # Specify 'cert' or 'key' to enable SSL / TLS mode
        # An encoding a line is going to be decoded from
        # - Pass '' (empty string) to prevent decoding
        "decode": "utf-8",
    }

    def __init__(
        self, app, pipeline, id=None, config=None, protocol_class=LineSourceProtocol
    ):
        """
        Description:

        """
        super().__init__(app, pipeline, id=id, config=config)

        self.Address = self.Config["address"]

        if "cert" in self.Config or "key" in self.Config:
            import bspump.asab.net

            sslbuilder = bspump.asab.net.SSLContextBuilder("[none]", config=self.Config)
            self.SSL = sslbuilder.build(protocol=ssl.PROTOCOL_SSLv23)
        else:
            self.SSL = None

        self.AcceptingSockets = []
        self.ConnectedClients = set()  # Set of active _client_connected_task()

        self.Protocol = protocol_class(app, pipeline, config=self.Config)

        app.PubSub.subscribe("Application.tick!", self._on_tick)

    def start(self, loop):
        """
        Description:

        """
        if self.Task is not None:
            return

        # Create all required sockets, bind them to specific ports and start listening
        for addrline in self.Address.split("\n"):
            addrline = addrline.strip()

            if addrline.count(":") == 1:
                host, port = self.Address.rsplit(":", maxsplit=1)

                addrinfo = socket.getaddrinfo(
                    host,
                    port,
                    family=socket.AF_UNSPEC,
                    type=socket.SOCK_STREAM,
                    flags=socket.AI_PASSIVE,
                )

                for family, socktype, proto, canonname, sockaddr in addrinfo:
                    s = socket.socket(family, socktype, proto)
                    try:
                        s.bind(sockaddr)
                    except OSError as e:
                        L.warning(
                            "Failed to start listening at '{}': {}".format(addrline, e)
                        )
                        continue

                    backlog = self.Config["backlog"]
                    if backlog == "":
                        s.listen()
                    else:
                        s.listen(int(backlog))

                    s.setblocking(False)
                    self.AcceptingSockets.append(s)

            else:
                self.Socket = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM)
                self.Socket.setblocking(False)
                self.Socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                self.Socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
                self.Socket.bind(self.Address)

                L.error("Invalid address specification: '{}'".format(addrline))

        super().start(loop)

    async def stop(self):
        """
        Description:

        """
        # Close client connections
        for t in self.ConnectedClients:
            t.cancel()
        self.ConnectedClients = set()

        # The main() will be canceled in the parent class
        await super().stop()

    async def main(self):
        """
        Description:

        """
        if len(self.AcceptingSockets) == 0:
            L.error("No listening socket configured")
            return

        await asyncio.gather(
            *[self._handle_accept(sock) for sock in self.AcceptingSockets],
            return_exceptions=True
        )

    async def _handle_accept(self, sock):
        """
        Description:

        """
        loop = self.Pipeline.App.Loop
        server_addr = sock.getsockname()
        while True:
            client_sock, client_addr = await loop.sock_accept(sock)
            t = loop.create_task(
                self._client_connected_task(client_sock, client_addr, server_addr)
            )
            self.ConnectedClients.add(t)

    async def _client_connected_task(self, client_sock, client_addr, server_addr):
        """
        Description:

        """
        client_sock.setblocking(False)

        if client_sock.family is socket.AF_INET:
            me = "{} {}".format(server_addr[0], server_addr[1])
            peer = "{} {}".format(client_addr[0], client_addr[1])
        elif client_sock.family is socket.AF_INET6:
            me = "{} {}".format(server_addr[0], server_addr[1])
            peer = "{} {}".format(client_addr[0], client_addr[1])
        else:
            me = server_addr
            peer = client_addr

        context = {
            "stream_type": client_sock.family.name,
            "stream_dir": "in",
            "stream_peer": peer,
            "stream_me": me,
        }
        if self.SSL is not None:
            stream = TLSStream(
                self.Pipeline.App.Loop, self.SSL, client_sock, server_side=True
            )
            ok = await stream.handshake()
            if not ok:
                return
        else:
            stream = Stream(self.Pipeline.App.Loop, client_sock)

        # This allows to send a reply to a client
        context["stream"] = stream

        inbound = self.Pipeline.App.Loop.create_task(
            self.Protocol.handle(self, stream, context)
        )
        outbound = self.Pipeline.App.Loop.create_task(stream.outbound())
        done, active = await asyncio.wait(
            {inbound, outbound}, return_when=asyncio.FIRST_COMPLETED
        )

        for t in active:
            # TODO: There could be outstanding data in the outbound queue
            #       Consider flushing them
            # Cancel remaining active tasks
            t.cancel()

        for t in done:
            try:
                await t
            except asyncio.CancelledError:
                pass
            except Exception:
                L.exception("Error when handling client socket")

        # Close the stream
        await stream.close()

    def _on_tick(self, event_name):
        # Remove clients that disconnected
        disconnected_client_tasks = [
            *filter(lambda task: task.done(), self.ConnectedClients)
        ]
        for task in disconnected_client_tasks:
            if task.done():
                self.ConnectedClients.remove(task)

                try:
                    task.result()
                except asyncio.CancelledError:
                    pass
                except Exception:
                    L.exception("Exception when handling client socket")



import csv
import logging
import os

from .fileabcsource import FileABCSource


L = logging.getLogger(__file__)


class FileCSVSource(FileABCSource):
    """
    Description:

    """

    ConfigDefaults = {
        "mode": "r",
        "newline": "",  # Required by CSV parser
        "dialect": "excel",
        "delimiter": ",",
        "doublequote": True,
        "escapechar": "",
        "lineterminator": os.linesep,
        "quotechar": '"',
        "quoting": csv.QUOTE_MINIMAL,
        "skipinitialspace": False,
        "strict": False,
    }

    def __init__(self, app, pipeline, fieldnames=None, id=None, config=None):
        super().__init__(app, pipeline, id=id, config=config)

        self.Dialect = csv.get_dialect(self.Config["dialect"])
        self.FieldNames = fieldnames

    def reader(self, f):
        """
        Description:

        **Parameters**

        f :

        :returns: ??

        |

        """
        kwargs = {}

        v = self.Config.get("delimiter")
        if v is not None:
            kwargs["delimiter"] = v

        v = self.Config.get("doublequote")
        if v is not None:
            kwargs["doublequote"] = v

        v = self.Config.get("escapechar")
        if v is not None:
            kwargs["escapechar"] = v

        v = self.Config.get("lineterminator")
        if v is not None:
            kwargs["lineterminator"] = v

        v = self.Config.get("quotechar")
        if v is not None:
            kwargs["quotechar"] = v

        v = self.Config.get("quoting")
        if v is not None:
            kwargs["quoting"] = v

        v = self.Config.get("skipinitialspace")
        if v is not None:
            kwargs["skipinitialspace"] = v

        v = self.Config.get("strict")
        if v is not None:
            kwargs["strict"] = v

        return csv.DictReader(
            f, dialect=self.Dialect, fieldnames=self.FieldNames, **kwargs
        )

    async def read(self, filename, f):
        """
        Description:

        **Parameters**

        filename :

        f :

        """

        for line in self.reader(f):
            await self.process(line, {"filename": filename})

            await self.simulate_event()



from ...abc import Expression

from ..value.valueexpr import VALUE


class WHEN(Expression):
    """
    Checks "if" condition passes - it is an `if` on steroids ;-)

    !WHEN
    - is:
            !EQ
            - !ITEM EVENT eggs
            - 2
    then: eggs

    - is:
            !LT
            - 9
            - !ITEM EVENT potatoes
            - 11
    then: potatoes

    - else:
            Nah

    """

    Attributes = False  # Filled during initialize() since attributes are dynamic

    Category = "Statements"

    def __init__(self, app, *, sequence):
        super().__init__(app)
        self.Items = sequence
        self.ItemsNormalized = []
        self.Attributes = {}
        self.OutletType = None  # Will be determined in `initialize()`
        self.Else = VALUE(self.App, value=None)

    def set(self, key, value):
        setattr(self, key, value)

        if "Test" in key:
            item_normalized = self.ItemsNormalized[int(key[4:])]
            self.ItemsNormalized[int(key[4:])] = (value, item_normalized[1])

        if "Then" in key:
            item_normalized = self.ItemsNormalized[int(key[4:])]
            self.ItemsNormalized[int(key[4:])] = (item_normalized[0], value)

    def initialize(self):
        item_counter = 0
        for n, i in enumerate(self.Items):
            # `test/then` branch
            if "test" in i and "then" in i:
                assert len(i) == 2

                vtest = i["test"]
                if not isinstance(vtest, Expression):
                    vtest = VALUE(self.App, value=vtest)

                attr_name = "Test{}".format(item_counter)
                setattr(self, attr_name, vtest)
                self.Attributes[attr_name] = [bool.__name__]

                vthen = i["then"]
                if not isinstance(vthen, Expression):
                    vthen = VALUE(self.App, value=vthen)

                attr_name = "Then{}".format(item_counter)
                setattr(self, attr_name, vthen)
                if self.OutletType is None:
                    self.OutletType = vthen.get_outlet_type()
                self.Attributes[attr_name] = self.OutletType

                self.ItemsNormalized.append((vtest, vthen))
                item_counter += 1

            # `else` branch
            elif "else" in i:
                assert len(i) == 1
                # TODO: Fix double-initialization when doing INCLUDE
                # assert('Else' not in self.Attributes)

                v = i["else"]
                if not isinstance(v, Expression):
                    v = VALUE(self.App, value=v)

                attr_name = "Else"
                setattr(self, attr_name, v)
                if self.OutletType is None:
                    self.OutletType = v.get_outlet_type()

                self.Attributes[attr_name] = self.OutletType

            else:
                raise RuntimeError("Unexpected items in '!WHEN': {}".format(i.keys()))

    def __call__(self, context, event, *args, **kwargs):
        for test, then in self.ItemsNormalized:
            res = test(context, event, *args, **kwargs)
            if res:
                return then(context, event, *args, **kwargs)

        return self.Else(context, event, *args, **kwargs)

    def get_outlet_type(self):
        return self.OutletType



import pytz
import datetime
import bspump.common
import bspump.unittest


class CustomTimeNormalizer(bspump.common.TimeZoneNormalizer):
    def process(self, context, event):
        native_time = event["@timestamp"]
        local_time = self.normalize(native_time)
        return local_time


class TestCustomTimeZoneNormalizer(bspump.unittest.ProcessorTestCase):
    def test_timezone_normalizer(self):
        events = [
            (
                None,
                {  # native time
                    "@timestamp": datetime.datetime(2000, 12, 30, 23, 59, 59, 59)
                },
            )
        ]

        self.set_up_processor(CustomTimeNormalizer)

        output = self.execute(events)

        self.assertEqual(  # local time represented in UTC
            [event for context, event in output],
            [datetime.datetime(2000, 12, 30, 22, 59, 59, 59, tzinfo=pytz.UTC)],
        )


class TestAbstractTimeZoneNormalizer(bspump.unittest.ProcessorTestCase):
    def test_abstract_timezone_normalizer(self):
        with self.assertRaises(TypeError):
            self.set_up_processor(bspump.common.TimeZoneNormalizer)



import logging

from ...abc import Expression
from ..value.valueexpr import VALUE
from ...declerror import DeclarationError

###

L = logging.getLogger(__name__)

###


class SUBSTRING(Expression):
    Attributes = {
        "Value": ["str"],
        "From": [""],  # TODO: This ...
        "To": ["*"],  # TODO: This ...
    }

    Category = "String"

    def __init__(self, app, *, arg_what, arg_from=0, arg_to=-1):
        super().__init__(app)
        self.Value = arg_what

        if not isinstance(arg_from, Expression):
            self.From = VALUE(app, value=arg_from)
        else:
            self.From = arg_from

        if not isinstance(arg_to, Expression):
            self.To = VALUE(app, value=arg_to)
        else:
            self.To = arg_to

    def get_outlet_type(self):
        return str.__name__

    def __call__(self, context, event, *args, **kwargs):
        try:
            _string = self.Value(context, event, *args, **kwargs)
            _from = self.From(context, event, *args, **kwargs)
            _to = self.To(context, event, *args, **kwargs)
            return _string[_from:_to]

        except Exception as e:
            raise DeclarationError(original_exception=e, location=self.get_location())



from ..abc.generator import Generator


class AnomalyAnalyzer(Generator):
    """
    AnomalyAnalyzer analyzes symptoms of anomalies to assign them to anomalies based on key
    dimensions within the symptom, which are defined in the "key_dimensions" configuration option
    for every symptom type.

    Every symptom should contain the following fields:
    {
            "@timestamp": UNIX_TIMESTAMP,
            "type": STRING,
            "key_dimension1": ...
            "key_dimension2": ...
            ...
            "other_dimension": ... (OPTIONAL)
            ...
            "status": future/open/closed (OPTIONAL)
    }
    """

    ConfigDefaults = {
        "key_dimensions": "default:user_id",  # Structure: type1:key_dimension1,key_dimension2;type2:key_dimension1 ...
    }

    def __init__(self, app, pipeline, id=None, config=None):
        super().__init__(app, pipeline, id, config)

        self.KeyDimensions = {}
        for key_value_str in self.Config["key_dimensions"].split(";"):
            key_value = key_value_str.split(":")
            self.KeyDimensions[key_value[0]] = key_value[1].split(",")

        metrics_service = app.get_service("asab.MetricsService")
        self.AnomalyAnalyzerCounter = metrics_service.create_counter(
            "anomaly.analyzer",
            init_values={
                "timestamp.miss": 0,
                "key.dimensions.miss": 0,
            },
        )

    def _construct_key(self, event, out_dimensions=None):
        key = ""

        # Obtain list of key dimensions that should be part of the event / symptom
        symptom_type = event.get("type", "")
        key_dimensions = self.KeyDimensions.get(symptom_type)
        if key_dimensions is None:
            key_dimensions = self.KeyDimensions.get("default")

        # Construct the key based on key dimensions
        for key_dimension in key_dimensions:
            if len(key_dimension) == 0:
                continue
            value = event.get(key_dimension)
            if value is None:
                continue
            key += str(value)
            if out_dimensions is not None:
                out_dimensions[key_dimension] = value

        if len(key) < 1:
            return None

        key += symptom_type

        return key

    async def generate(self, context, event, depth):
        # Obtain timestamp
        timestamp = event.get("@timestamp")
        if timestamp is None:
            self.AnomalyAnalyzerCounter.add("timestamp.miss", 1)
            return None

        if isinstance(timestamp, str):
            return None

        # Construct unique key for the anomaly
        key_dimensions = {}
        key = self._construct_key(event, key_dimensions)
        if key is None:
            self.AnomalyAnalyzerCounter.add("key.dimensions.miss", 1)
            return None

        # Pass the symptom to anomaly manager and its anomaly factory function
        # So far, the mapping is 1:1, i. e. one symptom type is associated with one anomaly type
        self.Pipeline.inject(
            context,
            {
                "@timestamp": timestamp,
                "key": key,
                "key_dimensions": key_dimensions,
                "type": event.get("type"),
                "symptom": event,
            },
            depth,
        )
